/*
 *  YASS 1.07
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "util.h"
#include "kword.h"
#include "tuple.h"
#include "prdyn.h"
#include "proba.h"
#include "align.h"
#include "assemble.h"
#include "display.h"
#include "global_var.h"


/*
 * Hach table division
 */
#define DIV_HACH 135
#define H_DIST(distance) ((distance)/DIV_HACH)

/*
 * Modulo access to the last tuple
 */

#ifdef MEMOPT
  /* query table index (little more time consuming but memory economic ) */
  int memopt_modmask = 0;
  #define MEMOPT_SETMOD   {int i=0;memopt_modmask = 0;for(i=0; ((keysize_query + gp_rho_stat + gp_delta_stat + 1) >= memopt_modmask);i++)  memopt_modmask |=  (1 << i);}
  #define MEMOPT_MOD      &(memopt_modmask)
  #define MEMOPT_TABSIZE   (memopt_modmask+1)


#else
  /* full table index (memory consuming) */ 
  #define MEMOPT_SETMOD
  #define MEMOPT_MOD       
  #define MEMOPT_TABSIZE   (keysize_query + keysize_text)
#endif

/*
 * Shift tables used in assemble algorithms
 */

int * deltashift = NULL;


int initialise_deltashift()
{
  int i,s=1;
  deltashift = (int * )malloc((3+2*gp_delta_stat)*sizeof(int));
  ASSERT(deltashift,"initialise_deltashift");
  for ( i=0 ; i<3+2*gp_delta_stat ; i++) {
    deltashift[i] = s * (i+1) / 2;
    s *= -1;
  }
  return 0;
}


int Assemble_Single(/*in*/ char * data, /*in*/ int datasize)
{
   

  /* main variables 
   * for the linked list computation
   */
  int * keylist;
  int * first;
  int keysize_query = datasize - gp_seed_span + 1;
  register int index=0;
  register int index_fin=0;
  
  /* for the linking algorithm */
  int i_previous;
  int j;
  int distance,observed_distance,delta_distance;

  /* Hach table */
  int j_hack; 
  register int borne_min_hash=gp_rho_stat/2-1;
  register int borne_max_hash=DIV_HACH-gp_rho_stat/2+1;
  int *last_tuple_pos_with_dist_hach;

  /* to keep lists of tuple found */
  tuplelist * tuple_lists = CreateTupleList();
  
   
  /* auxiliary variables */
  int code;
  int  *   last_tuple_pos_with_dist = NULL;
  tuple ** last_tuple_ptr_with_dist     = NULL;
 

  /* sequence covering variables */ 
  tuple     * t, * t_;
  gv_last_tl =  tuple_lists;

  /* [0] left correction factor set to zero */
  gv_left_correction =  0;

  /* [1] key linked list creation */
  CreateKeyList(data,datasize,&first,&keylist);
  
  /* [2] allocations */

  last_tuple_pos_with_dist = (int *) malloc((unsigned)(keysize_query * sizeof(int)));
  ASSERT(last_tuple_pos_with_dist,"Assemble");
  memset(last_tuple_pos_with_dist,0xff,(unsigned)(keysize_query * sizeof(int)));
  

  last_tuple_ptr_with_dist = (tuple **) malloc((unsigned)(keysize_query * sizeof(tuple *)));
  ASSERT(last_tuple_ptr_with_dist,"Assemble");
  memset(last_tuple_ptr_with_dist,0x00,(unsigned)(keysize_query * sizeof(tuple *)));

  last_tuple_pos_with_dist_hach = (int *) malloc((unsigned)((keysize_query/DIV_HACH+1) * sizeof(int)));
  ASSERT(last_tuple_pos_with_dist_hach,"Assemble");
  memset(last_tuple_pos_with_dist_hach,0xff,(unsigned)((keysize_query/DIV_HACH+1) * sizeof(int)));


  

  /* [3] algorithm to create linked tuples */

#ifdef STATS
  last_clock = clock();
#endif
  
  /* for each kwork of the Text */
  for(gv_i_current = 0 ; gv_i_current < keysize_query ; gv_i_current ++ ){

    code      = Key(data,gv_i_current); /* we take the code of the kmot 'w' located at gv_i_current position     */
    index     = first[code];  /* search the first occurrence of 'w' on the sequence */
    index_fin = first[code+1];
    i_previous= keylist[index];

    /* consider all the occurrences of 'w' before gv_i_current */
    while(i_previous < gv_i_current - gp_distdiag && index<index_fin ){
      
      distance = gv_i_current - i_previous;
      /* consider all the interesting distances (diagonals) by adding indel bias  0 +1 -1 +2... until we reach a new tuple or the upper indel bound */
      observed_distance = distance;
      delta_distance = 0;

      /* Search on the hach table if there are one diagonale */
      j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance)]; 
      code=((j_hack< gv_i_current - gp_rho_stat)||(j_hack > gv_i_current));
      if (observed_distance%DIV_HACH>borne_max_hash) {
	j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance)+1]; 
	code=code&&((j_hack< gv_i_current - gp_rho_stat)||(j_hack > gv_i_current)); 
      }
      if (observed_distance%DIV_HACH<borne_min_hash) {
	j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance)-1]; 
	code=code&&((j_hack< gv_i_current - gp_rho_stat)||(j_hack > gv_i_current));
      }
      if (code==0) 
       while( delta_distance <= 2 * gp_delta_stat + 1){
	
	/* search for interesting tuples */
	j = last_tuple_pos_with_dist[observed_distance];
	
	/* if tuples found are respecting criteria (neighboors on diagonal and on sequences) */
	if ( j > gv_i_current - gp_rho_stat                  &&
	     j < gv_i_current                                &&
	     j - observed_distance < gv_i_current - distance &&
	     j != -1 ){
	  
	  /*---------------------------------------------------------------------*/
	  
	  /* (1) possibly create one instance of tuple [j]
	   * if this one was not in any list.
	   */
	  
	  if (last_tuple_ptr_with_dist[observed_distance] == NULL){
            
#ifdef STATS
	    nb_chains_built++;
#endif

	    /* create at tuple list element */
	    gv_last_tl->next = CreateTupleList();
	    gv_last_tl = gv_last_tl->next;
	    /* create a tuple [j] and update lists */
	    t  = CreateTuple(j,observed_distance);
	    
	    gv_last_tl->first_tuple = t;
	    last_tuple_ptr_with_dist[observed_distance] = t;
	  }

	  /* (2) we search for the last tuple [j] instanciation
	   */
	  t = last_tuple_ptr_with_dist[observed_distance];
	  	
	  /*
	   * (3) iterate tuples and add this new one to the most adapted one
	   */

	  do{
	    
	    if (  distance == t->distance  && gv_i_current - t->occurrence <= gp_seed_span ){
	      
	      /* (3a) possible fusion with this tuple (no instanciation)
	       *
	       * [inter-tuple overlap]
	       */
	      
	      t->size        +=  gv_i_current -  t->occurrence;
	      t->overlappedhits   <<= (gv_i_current -  t->occurrence - 1);
	      t->overlappedhits    &= AUTOMATAMASK;
	      t->bitweight         += gp_automata[t->overlappedhits];
	      t->occurrence    =  gv_i_current;
	      t->overlappedhits   <<= 1;
	      t->overlappedhits    |= 1;
	      
	      goto no_maj_last_tuple_ptr_with_dist;
	  
	    }
	    
	    /* list searching for the most fitted */
	    if(!(t->next))
	      break;
	    t = t->next;
	    
	  }while(1);
	  
	  /*
	   * (3d) any fusion has not been possible
	   *
	   * we instanciate the tuple and add this one at the end of the chain.
	   *
	   */

	  t_ = CreateTuple(gv_i_current, distance);

	  t->next = t_;
	  last_tuple_ptr_with_dist[distance] = t_;
	  /*
	   * (4) dont reset last_tuple_ptr_with_dist[distance] to NULL
	   * because it has been modified in (3)
	   */
	  goto no_maj_last_tuple_ptr_with_dist;
	  
	}
	
	/*---------------------------------------------------------------------*/
	

	/* diagonal walk d,d+1d-1,... */
	delta_distance++;
	observed_distance = distance + deltashift[delta_distance];
	observed_distance = MAX(MIN(observed_distance,keysize_query-1),1);
	/* border effects avoided*/
	  
      }/* while : diagonal walk */
	

      /* reset last tuple instancied to NULL (because not instance is created)
       */

      last_tuple_ptr_with_dist[distance] = NULL;
no_maj_last_tuple_ptr_with_dist:
      /*
       * update last tuple position  with diagonal = 'distance'
       */
#ifdef STATS
      nb_seeds++;
#endif
      last_tuple_pos_with_dist[distance] = gv_i_current;
      last_tuple_pos_with_dist_hach[H_DIST(distance)]=gv_i_current;
      /* iterate 'w' code list before gv_i_current position */
      index ++;
      i_previous = keylist[index]; 
    }     
    if (!(gv_i_current % ALIGN_EVERY_NB_ITER)){
#ifdef TRACE
      Display_Progress((unsigned)gv_i_current,(unsigned)(keysize_query - 1));
#endif
#ifdef STATS
      clock_chain += clock() - last_clock;
      last_clock=  clock();
#endif		
#ifdef DEBUGASSEMBLE
      DisplayListTupleList(tuple_lists);
#endif
      AlignAndFree(tuple_lists, 
		   data, datasize,
		   data, datasize,
		   0,0);
#ifdef STATS
      clock_align += clock() - last_clock;
      last_clock=  clock();
#endif
    }
    
  }/*end of [3] */
#ifdef TRACE
      Display_Progress((unsigned)(unsigned)gv_i_current,keysize_query - 1);
#endif  
 
#ifdef STATS
  clock_chain += clock() - last_clock;
  last_clock=  clock();
#endif	

  /* on aligne les tuples restants */
#ifdef DEBUGASSEMBLE
  DisplayListTupleList(tuple_lists);
#endif
  AlignAndFree(tuple_lists, 
	       data, datasize,
	       data, datasize,
	       1,0);
#ifdef STATS	
   clock_align += clock() - last_clock;
   last_clock =  clock();
#endif
   free(last_tuple_pos_with_dist);
   free(last_tuple_ptr_with_dist);
   free(keylist);
   free(first);
   return  0;
}


int Assemble_SingleRev(/*in*/ char * datarev, /*in*/ char * data, /*in*/ int datasize)
{
   

  /* main variables */
  /* for the linked list computation */
  int * keylist;
  int * first;
  int keysize_query = datasize - gp_seed_span + 1;
#ifndef MEMOPT	
  int keysize_text  = keysize_query ; 
#endif	
  /* for the linking algorithm */
  int index,index_fin;
  int i_previous;
  int j;
  int distance,observed_distance,delta_distance;

  /* to keep lists of tuple found */
  tuplelist * tuple_lists = CreateTupleList();

  /* Hach table */
  int j_hack; 
  register int borne_min_hash=gp_rho_stat/2-1;
  register int borne_max_hash=DIV_HACH-gp_rho_stat/2+1;
  int *last_tuple_pos_with_dist_hach;  
   

  /* auxiliary variables */
  int code;
  int  *   last_tuple_pos_with_dist = NULL;
  tuple ** last_tuple_ptr_with_dist = NULL;
 

  /* sequence covering variables */ 
  tuple     * t, * t_;
  gv_last_tl =  tuple_lists;

  /* [0] left correction factor set to keysize_query */
  gv_left_correction =  keysize_query;
  MEMOPT_SETMOD;  
 
  /* [1] key linked list creation */
  CreateKeyList(datarev,datasize,&first,&keylist);

 
  
  /* [2] allocations */
  last_tuple_pos_with_dist = (int *) malloc((unsigned)( MEMOPT_TABSIZE * sizeof(int)));
  ASSERT(last_tuple_pos_with_dist,"Assemble");
  memset(last_tuple_pos_with_dist,0xff,(unsigned)( MEMOPT_TABSIZE * sizeof(int)));
  

  last_tuple_ptr_with_dist = (tuple **) malloc((unsigned)( MEMOPT_TABSIZE * sizeof(tuple *)));
  ASSERT(last_tuple_ptr_with_dist,"Assemble");
  memset(last_tuple_ptr_with_dist,0x00,(unsigned)( MEMOPT_TABSIZE *sizeof(tuple *)));

  last_tuple_pos_with_dist_hach = (int *) malloc((unsigned)(((keysize_query+keysize_query)/DIV_HACH+1) * sizeof(int)));
  ASSERT(last_tuple_pos_with_dist_hach,"Assemble");
  memset(last_tuple_pos_with_dist_hach,0xff,(unsigned)(((keysize_query+keysize_query)/DIV_HACH+1) * sizeof(int)));


  /* [3] algorithm to create linked tuples */

#ifdef STATS
  last_clock = clock();
#endif
  
  /* for each kwork of the Text */
  for(gv_i_current = 0 ; gv_i_current < keysize_query ; gv_i_current ++ ){
    
   
    code      = Key(data, gv_i_current); /* we take the code of the kmot 'w' located at gv_i_current position     */
    index     = first[code];  /* search the first occurrence of 'w' on the sequence */
    index_fin = first[code+1];
    i_previous= keylist[index];

    
    /* consider all the occurrences of 'w' before gv_i_current */
   
    
    while(i_previous < keysize_query - gv_i_current  && index<index_fin){  

      distance =   gv_i_current - i_previous + keysize_query ;      
      /* consider all the interesting distances (diagonals) by adding indel bias  0 +1 -1 +2... until we reach a new tuple or the upper indel bound */
      observed_distance = distance;
      delta_distance = 0;
     
      j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance )]; 
      code=((j_hack< gv_i_current - gp_rho_stat)||(j_hack > gv_i_current));
      if (observed_distance%DIV_HACH>borne_max_hash) {
	j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance)+1]; 
	code=code&&((j_hack< gv_i_current - gp_rho_stat)||(j_hack > gv_i_current)); 
      }
      if (observed_distance%DIV_HACH<borne_min_hash) {
	j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance)-1]; 
	code=code&&((j_hack< gv_i_current - gp_rho_stat)||(j_hack > gv_i_current));
      }
	
      if (code==0)
      while( delta_distance <= 2 * gp_delta_stat + 1){
	
	/* search for interesting tuples */
	j = last_tuple_pos_with_dist[observed_distance MEMOPT_MOD ];

	/* Search on the hach table if there are one diagonale */
	/* if tuples found are respecting criteria (neighboors on diagonal and on sequences) */
	if ( j > gv_i_current - gp_rho_stat                  &&                                         
	     j < gv_i_current                                && 
	     j - observed_distance < gv_i_current - distance && 
	     j != -1 ){ 
	  
	  
	  /*---------------------------------------------------------------------*/
	  
	  /* (1) possibly create one instance of tuple [j] 
	   * if this one was not in any list. 
	   */
	  
	  if (last_tuple_ptr_with_dist[observed_distance MEMOPT_MOD] == NULL){
            
#ifdef STATS
	    nb_chains_built++;
#endif

	    /* create at tuple list element */
	    gv_last_tl->next = CreateTupleList();
	    gv_last_tl = gv_last_tl->next;
	    
	    /* create a tuple [j] and update lists */
	    t  = CreateTuple(j,observed_distance);

	    
	    gv_last_tl->first_tuple = t;
	    last_tuple_ptr_with_dist[observed_distance MEMOPT_MOD] = t;
	  }

	  /* (2) we search for the last tuple [j] instanciation 
	   */
	  t = last_tuple_ptr_with_dist[observed_distance MEMOPT_MOD];
	  	
	  /*
	   * (3) iterate tuples and add this new one to the most adapted one
	   */

	  do{
	    
	    if (  distance == t->distance  && gv_i_current - t->occurrence <= gp_seed_span ){
	      
	      /* (3a) possible fusion with this tuple (no instanciation)
	       *	
	       * [inter-tuple overlap]	
	       */      
	      t->size        +=  gv_i_current -  t->occurrence;
	      t->overlappedhits   <<= (gv_i_current -  t->occurrence - 1);
	      t->overlappedhits    &= AUTOMATAMASK;
	      t->bitweight         += gp_automata[t->overlappedhits];
	      t->occurrence    =  gv_i_current;
	      t->overlappedhits   <<= 1;
	      t->overlappedhits    |= 1;
	      
	      goto no_maj_last_tuple_ptr_with_dist;
	  
	    }
	    
	    /*	parcours de la liste */
	    if(!(t->next))
	      break;
	    t = t->next;
	    
	  }while(1);
	  
	  /*
	   * (3d) any fusion has not been possible
	   *
	   * we instanciate the tuple and add this one at the end of the chain.
	   *
	   */
	  
	  t_ = CreateTuple(gv_i_current, distance);
	  t->next = t_;
	  last_tuple_ptr_with_dist[distance MEMOPT_MOD] = t_;
	  
	  /*
	   * (4) dont reset last_tuple_ptr_with_dist[distance] to NULL 
	   * because it has been modified in (3)
	   */
	  goto no_maj_last_tuple_ptr_with_dist;
	  
	}
	
	/*---------------------------------------------------------------------*/
	

	/* diagonal walk d,d+1,d-1,... */
	delta_distance ++;
	observed_distance = distance + deltashift[delta_distance];
	observed_distance = MAX(MIN(observed_distance,keysize_query-1),1);
	/* border effects avoided*/
	  
      }/*  while : diagonal walk */
	

      /* reset last tuple instancied to NULL (because not instance is created) 
       */ 
	
      last_tuple_ptr_with_dist[distance MEMOPT_MOD] = NULL;
	    
no_maj_last_tuple_ptr_with_dist:

      /* 
       * update last tuple position  with diagonal = 'distance'
       */
#ifdef STATS
      nb_seeds++;
#endif
      last_tuple_pos_with_dist_hach[H_DIST(distance)]=gv_i_current;
      last_tuple_pos_with_dist[distance MEMOPT_MOD] = gv_i_current;

      /* iterate 'w' code list before gv_i_current position */
      index ++;
      i_previous = keylist[index];
    }  
    if (!(gv_i_current % ALIGN_EVERY_NB_ITER)){
#ifdef TRACE
      Display_Progress((unsigned)gv_i_current,(unsigned)(keysize_query - 1));
#endif
#ifdef STATS
      clock_chain += clock() - last_clock;
      last_clock=  clock();
#endif		
#ifdef DEBUGASSEMBLE
      DisplayListTupleList(tuple_lists);
#endif
      AlignAndFree(tuple_lists, 
		   datarev, datasize,
		   data,    datasize,
		   0,0);

#ifdef STATS
      clock_align += clock() - last_clock;
      last_clock=  clock();
#endif
    }
    
  }/*end of [3] */
#ifdef TRACE
  Display_Progress((unsigned)gv_i_current,(unsigned)(keysize_query - 1));
#endif  
 
#ifdef STATS
  clock_chain += clock() - last_clock;
  last_clock=  clock();
#endif	


  /* on aligne les tuples restants */
#ifdef DEBUGASSEMBLE
  DisplayListTupleList(tuple_lists);
#endif
  AlignAndFree(tuple_lists, 
	       datarev,    datasize,
	       data, datasize,
	       1,0);
#ifdef STATS	
   clock_align += clock() - last_clock;
   last_clock =  clock();
#endif
   free(last_tuple_pos_with_dist);
   free(last_tuple_ptr_with_dist);
   free(keylist);
   free(first);
   return  0;
}









int Assemble_Double( char * data_query, int datasize_query,
		     char * data_text,  int datasize_text )
{
  
  
  /* main variables */
  /* for the linked list computation */
  int * keylist_query;
  int * first_query;
  int keysize_query = datasize_query - gp_seed_span + 1;
  int keysize_text  = datasize_text  - gp_seed_span + 1;
  
  /* for the linking algorithm */
  int i_previous;
  int j;
  int distance,observed_distance,delta_distance;
  
  /* to keep lists of tuple found */
  tuplelist * tuple_lists = CreateTupleList();

  /* Hach table */
  int j_hack; 
  register int borne_min_hash=gp_rho_stat/2-1;
  register int borne_max_hash=DIV_HACH-gp_rho_stat/2+1;
  int *last_tuple_pos_with_dist_hach;

  
  /* auxiliary variables */
  int  index,index_fin;
  int code;
  int  *   last_tuple_pos_with_dist = NULL;
  tuple ** last_tuple_ptr_with_dist = NULL;
 

  /* sequence covering variables */ 
  tuple     * t, * t_;
  gv_last_tl =  tuple_lists;
  
  /* [0] left correction factor set to keysize_query */
  gv_left_correction =  keysize_query;
  MEMOPT_SETMOD;

  /* [1] key linked list creation */
  CreateKeyList(data_query,datasize_query,&first_query,&keylist_query);
  
  /* [2] allocations */
  last_tuple_pos_with_dist = (int *) malloc((unsigned)( MEMOPT_TABSIZE * sizeof(int)));
  ASSERT(last_tuple_pos_with_dist,"Assemble");
  memset(last_tuple_pos_with_dist,0xff,(unsigned)( MEMOPT_TABSIZE * sizeof(int)));
    

  last_tuple_ptr_with_dist = (tuple **) malloc((unsigned)( MEMOPT_TABSIZE * sizeof(tuple *)));
  ASSERT(last_tuple_ptr_with_dist,"Assemble");
  memset(last_tuple_ptr_with_dist,0x00,(unsigned)( MEMOPT_TABSIZE * sizeof(tuple *)));

  last_tuple_pos_with_dist_hach = (int *) malloc((unsigned)(((keysize_query+keysize_query)/DIV_HACH+1) * sizeof(int)));
  ASSERT(last_tuple_pos_with_dist_hach,"Assemble");
  memset(last_tuple_pos_with_dist_hach,0xff,(unsigned)(((keysize_query+keysize_query)/DIV_HACH+1) * sizeof(int)));

  
  /* [3] algorithm to create linked tuples */

#ifdef STATS
  last_clock = clock();
#endif
  
  /* for each kwork of the Text */
  for(gv_i_current = 0 ; gv_i_current < keysize_text ; gv_i_current ++ ){
    
    code      = Key(data_text,gv_i_current);  /* we take the code of the kmot 'w' located at gv_i_current position on the "Text" sequence */
    index     = first_query[code];  /* search the first occurrence of 'w' on the sequence */
    index_fin = first_query[code+1];
    i_previous= keylist_query[index];       /* search the first occurrence of 'w' on the sequence "Query" */
    
    /* for each occurrence of 'w' on the "Query" sequence */
    while(index<index_fin){ 
      
      distance =  gv_i_current + keysize_query - i_previous;
      /* consider all the interesting distances (diagonals) by adding indel bias  0 +1 -1 +2... until we reach a new tuple or the upper indel bound */
      observed_distance = distance;
      delta_distance = 0;

      j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance )]; 
      code=((j_hack< gv_i_current - gp_rho_stat)||(j_hack > gv_i_current));
      if (observed_distance%DIV_HACH>borne_max_hash) {
	j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance)+1]; 
	code=code&&((j_hack< gv_i_current - gp_rho_stat)||(j_hack > gv_i_current)); 
      }
      if (observed_distance%DIV_HACH<borne_min_hash) {
	j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance)-1]; 
	code=code&&((j_hack< gv_i_current - gp_rho_stat)||(j_hack > gv_i_current));
      }
	
      if (code==0)
      while( delta_distance <= 2 * gp_delta_stat + 1 ){
		
	/* search for interesting tuples */
	j = last_tuple_pos_with_dist[observed_distance MEMOPT_MOD];
	
	/* if tuples found are respecting criteria (neighboors on diagonal and on sequences) */
	if (j > gv_i_current -  gp_rho_stat                  && 
	    j < gv_i_current                                 &&
	    j - observed_distance < gv_i_current - distance  && 
	    j != -1 ) { 
	  
	  
	  /*---------------------------------------------------------------------*/
	  
	  /* (1) possibly create one instance of tuple [j] 
	   * if this one was not in any list. 
	   */
	  
	  if (last_tuple_ptr_with_dist[observed_distance MEMOPT_MOD] == NULL){
	    
#ifdef STATS
	    nb_chains_built++;
#endif
	    /* create at tuple list element */
	    gv_last_tl->next = CreateTupleList();
	    gv_last_tl = gv_last_tl->next;
	    
	    /* create a tuple [j] and update lists */
	    t = CreateTuple(j,observed_distance);
	    gv_last_tl->first_tuple = t;
	    last_tuple_ptr_with_dist[observed_distance MEMOPT_MOD] = t;
	    
	  }
	  
	 
	  /* (2) we search for the last tuple [j] instanciation 
	   */
	  t = last_tuple_ptr_with_dist[observed_distance MEMOPT_MOD];
	  	
	  
	  /*
	   * (3) iterate tuples and add this new one to the most adapted one
	   */

	  do{
	    
	    if (  distance == t->distance  && gv_i_current - t->occurrence <= gp_seed_span ){
	      
	      /* (3a) possible fusion with this tuple (no instanciation)
	       *	
	       * [inter-tuple overlap]	
	       */ 
	     
	      t->size        +=  gv_i_current -  t->occurrence;  
	      t->overlappedhits   <<= (gv_i_current -  t->occurrence - 1);
	      t->overlappedhits    &= AUTOMATAMASK;
	      t->bitweight         += gp_automata[t->overlappedhits];
	      t->occurrence    =  gv_i_current; 
	      t->overlappedhits   <<= 1;
	      t->overlappedhits    |= 1;
	      
	      goto no_maj_last_tuple_ptr_with_dist;
	  
	    }
	    	  
	    /*	parcours de la liste */
	    if(!(t->next))
	      break;
	    t = t->next;
	  
	  }while(1);
	  
	  /*
	   * (3d) any fusion has not been possible
	   *
	   * we instanciate the tuple and add this one at the end of the chain.
	   *
	   */
	  
	  t_ = CreateTuple(gv_i_current, distance);
	  t->next = t_;
	  last_tuple_ptr_with_dist[distance MEMOPT_MOD] = t_;
	  
	  /*
	   * (4) dont reset last_tuple_ptr_with_dist[distance] to NULL
	   * because it has been modified in (3)
	   */
	  goto no_maj_last_tuple_ptr_with_dist;
	  
	}
	
	/*---------------------------------------------------------------------*/
	

	/* diagonal walk d,d+1,d-1,... */
	delta_distance ++;
	observed_distance = distance + deltashift[delta_distance];
	observed_distance = MAX(MIN(observed_distance,gv_i_current+keysize_query),gv_i_current+1);
	/* border effects avoided*/
	  
      }/*  while : diagonal walk */
	

      /* reset last tuple instancied to NULL (because not instance is created)
       */ 
	
      last_tuple_ptr_with_dist[distance MEMOPT_MOD] = NULL;
	    
no_maj_last_tuple_ptr_with_dist:

      /* 
       * update last tuple position  with diagonal = 'distance'
       */
#ifdef STATS
      nb_seeds++;
#endif
      last_tuple_pos_with_dist[distance MEMOPT_MOD] = gv_i_current;	  
      last_tuple_pos_with_dist_hach[H_DIST(distance)]=gv_i_current;

      /* iterate 'w' code list before gv_i_current position */
      index ++;
      i_previous= keylist_query[index]; 
    }  
    if (!(gv_i_current % ALIGN_EVERY_NB_ITER)){
#ifdef TRACE
      Display_Progress((unsigned)gv_i_current,(unsigned)(datasize_text - 1));    
#endif
#ifdef STATS
      clock_chain += clock() - last_clock;
      last_clock=  clock();
#endif
#ifdef DEBUGASSEMBLE
      DisplayListTupleList(tuple_lists);
#endif	
      AlignAndFree(tuple_lists,
		   data_query,  datasize_query,
		   data_text, datasize_text, 
		   0,0);
#ifdef STATS
      clock_align += clock() - last_clock;
      last_clock   = clock();
#endif 

    }
    
  }/*end of [3] */
#ifdef TRACE
  Display_Progress((unsigned)gv_i_current,(unsigned)(datasize_text - 1));    
#endif
#ifdef STATS
  clock_chain += clock() - last_clock;
  last_clock=  clock();
#endif	
#ifdef DEBUGASSEMBLE
  DisplayListTupleList(tuple_lists);
#endif	
  AlignAndFree(tuple_lists,
	       data_query,  datasize_query,
	       data_text, datasize_text, 
	       1,0);
#ifdef STATS		
   clock_align += clock() - last_clock;
   last_clock =  clock();
#endif   
   free(last_tuple_pos_with_dist);
   free(last_tuple_ptr_with_dist);
   free(keylist_query);
   free(first_query);
   return  0;
}








int MultiAssemble_Double(char * data_query, int datasize_query,
			 char * data_text,  int datasize_text, 
			 int nbchunks_text, char ** chunkname_text, int * chunksize_text,int * chunkstrt_text)
{
  /* used for chunk selection */
  int i_chunk                    = 0;
  int i_base                     = 0;
  int max_chunk_datasize_text    = 0;
  int keysize_text     = 0;
  
  int     chunk_keysize_text     = 0;
  char *  chunk_data_text        = data_text;
  
  
  /*  main variables 
   *  for the linked list computation
   */
  
  int * keylist_query;
  int * first_query;
  int   keysize_query = datasize_query - gp_seed_span + 1;
  
  /* for the linking algorithm */
  int i_previous;
  int index,index_fin;
  int j;
  int distance,observed_distance,delta_distance;
  
  /* to keep lists of tuple found */
  tuplelist * tuple_lists = CreateTupleList();

  
  /* auxiliary variables */
  int      code;
  int  *   last_tuple_pos_with_dist = NULL;
  tuple ** last_tuple_ptr_with_dist = NULL;
 

  /* sequence covering variables */ 
  tuple     * t, * t_;
  
  gv_last_tl =  tuple_lists;
  
  for(i_chunk = 0 ; i_chunk < nbchunks_text; i_chunk++)
    max_chunk_datasize_text = MAX(max_chunk_datasize_text, chunksize_text[i_chunk]);
  keysize_text = max_chunk_datasize_text - gp_seed_span + 1; 


  /* [0] left correction factor set to keysize_query */
  gv_left_correction =  keysize_query;
  MEMOPT_SETMOD;

  /* [1] key linked list creation */
  CreateKeyList(data_query,datasize_query,&first_query,&keylist_query);
  
  

  /* [2] allocations */
  last_tuple_pos_with_dist = (int *) malloc((unsigned)(MEMOPT_TABSIZE * sizeof(int)));
  ASSERT(last_tuple_pos_with_dist,"Assemble");
   
  last_tuple_ptr_with_dist = (tuple **) malloc((unsigned)(MEMOPT_TABSIZE * sizeof(tuple *)));
  ASSERT(last_tuple_ptr_with_dist,"Assemble");
 

#ifdef STATS
  last_clock = clock();
#endif
  

  /* for each chunk element */
  for( i_chunk = 0 ; i_chunk < nbchunks_text ; i_chunk++ ){
    
    /*  
     * [3] clean used tables  and
     *     and set parameters  
     */
    
    
    memset(last_tuple_pos_with_dist,0xff,(unsigned)( MEMOPT_TABSIZE * sizeof(int)));
    memset(last_tuple_ptr_with_dist,0x00,(unsigned)( MEMOPT_TABSIZE * sizeof(tuple *)));
    chunk_keysize_text  =  chunksize_text[i_chunk] - gp_seed_span + 1;
    chunk_data_text    +=  ((i_chunk == 0) ? (0) : (chunksize_text[i_chunk-1]));
    i_base             +=  ((i_chunk == 0) ? (0) : (chunksize_text[i_chunk-1])); 
    
    /* [4] algorithme de creation des tuples chains */

    /* for each text kword  */
    for(gv_i_current = 0 ; gv_i_current < chunk_keysize_text ; gv_i_current ++ ){
     
      code      = Key(chunk_data_text,gv_i_current);    /* we take the code of the kmot 'w' located at gv_i_current position on the "Text" */
      index=first_query[code];     /* search the first occurrence of 'w' on the sequence "Query" */
      index_fin=first_query[code+1];
      i_previous=keylist_query[index];
      
      /* for each occurrence of 'w' on the "Query" sequence */
      while(index<index_fin){ 
	
	distance =  gv_i_current + keysize_query - i_previous;
	/* consider all the interesting distances (diagonals) by adding indel bias  0 +1 -1 +2... until we reach a new tuple or the upper indel bound */
	observed_distance = distance;
	delta_distance = 0;
	
	while( delta_distance <= 2 * gp_delta_stat + 1 ){
	  
	  /* search for interesting tuples */
	  j = last_tuple_pos_with_dist[observed_distance MEMOPT_MOD];
	  
	  /* if tuples found are respecting criteria (neighboors on diagonal and on sequences) */
	  if (j > gv_i_current -  gp_rho_stat                  && 
	      j < gv_i_current                                 &&
	      j - observed_distance < gv_i_current - distance  && 
	      j != -1 ) { 
	    
	    
	    
	    /*---------------------------------------------------------------------*/
	    
	    /* (1) possibly create one instance of tuple [j] 
	     * if this one was not in any list. 
	     */
	    
	    if (last_tuple_ptr_with_dist[observed_distance MEMOPT_MOD] == NULL){
	      
#ifdef STATS
	      nb_chains_built++;
#endif
	      /* create at tuple list element */
	      gv_last_tl->next = CreateTupleList();
	      gv_last_tl = gv_last_tl->next;
	      
	      /* create a tuple [j] and update lists */
	      t = CreateTuple(j,observed_distance);
	      gv_last_tl->first_tuple = t;
	      last_tuple_ptr_with_dist[observed_distance MEMOPT_MOD] = t;
	      
	    }
	    
	    
	    /* (2) we search for the last tuple [j] instanciation 
	     */
	    t = last_tuple_ptr_with_dist[observed_distance MEMOPT_MOD];
	    
	    
	    /*
	     * (3) iterate tuples and add this new one to the most adapted one
	     */
	    
	    do{
	      
	      if (  distance == t->distance  && gv_i_current - t->occurrence <= gp_seed_span ){
		
		/* (3a) possible fusion with this tuple (no instanciation)
		 *	
		 * [inter-tuple overlap]	
		 */ 
		
		t->size        +=  gv_i_current -  t->occurrence;  
		t->overlappedhits   <<= (gv_i_current -  t->occurrence - 1);
		t->overlappedhits    &= AUTOMATAMASK;
		t->bitweight         += gp_automata[t->overlappedhits];
		t->occurrence    =  gv_i_current; 
		t->overlappedhits   <<= 1;
		t->overlappedhits    |= 1;
		
		goto no_maj_last_tuple_ptr_with_dist;
		
	      }
	      
	      
	      /*	parcours de la liste */
	      if(!(t->next))
		break;
	      t = t->next;
	      
	    }while(1);
	    
	    /*
	     * (3d) any fusion has not been possible
	     *
	     * we instanciate the tuple and add this one at the end of the chain.
	     *
	     */
	    
	    t_ = CreateTuple(gv_i_current, distance);
	    t->next = t_;
	    last_tuple_ptr_with_dist[distance MEMOPT_MOD] = t_;
	    
	    /*
	     * (4) dont reset last_tuple_ptr_with_dist[distance] to NULL
	     * because it has been modified in (3)
	     */
	    goto no_maj_last_tuple_ptr_with_dist;
	    
	  }
	  
	  /*---------------------------------------------------------------------*/
	  
	  
	  /* diagonal walk d,d+1,d-1,... */
	  delta_distance ++;
	  observed_distance = distance + deltashift[delta_distance];
	  observed_distance = MAX(MIN(observed_distance,gv_i_current+keysize_query),gv_i_current+1);
	  /* border effects avoided*/
	  
	}/*  while : diagonal walk */
	
	
	/* reset last tuple instancied to NULL (because not instance is created)
	 */ 
	
	last_tuple_ptr_with_dist[distance MEMOPT_MOD] = NULL;
	
      no_maj_last_tuple_ptr_with_dist:
	
	/* 
	 * update last tuple position  with diagonal = 'distance' 
	 */
#ifdef STATS
	nb_seeds++;
#endif
	last_tuple_pos_with_dist[distance MEMOPT_MOD] = gv_i_current;	  
	/* iterate 'w' code list before gv_i_current position */
	index++;
        i_previous=keylist_query[index];
      }  
      
      if (!(gv_i_current % ALIGN_EVERY_NB_ITER)){
#ifdef TRACE
	Display_Progress((unsigned)gv_i_current+i_base,(unsigned)(datasize_text-1));    
#endif
#ifdef STATS
	clock_chain += clock() - last_clock;
	last_clock=  clock();
#endif
#ifdef DEBUGASSEMBLE
	DisplayListTupleList(tuple_lists);
#endif	
	AlignAndFree(tuple_lists,
		     data_query,        datasize_query,
		     chunk_data_text,   chunksize_text[i_chunk], 
		     0,i_chunk);
#ifdef STATS
	clock_align += clock() - last_clock;
	last_clock   = clock();
#endif 
	
      }
    }/*end of  [3] (keysize_text loop)  */
#ifdef TRACE
    Display_Progress((unsigned)gv_i_current+i_base,(unsigned)(datasize_text));   
#endif
#ifdef STATS
    clock_chain += clock() - last_clock;
    last_clock=  clock();
#endif	
#ifdef DEBUGASSEMBLE
    DisplayListTupleList(tuple_lists);
#endif	
    
    AlignAndFree(tuple_lists,
		 data_query     ,  datasize_query,
		 chunk_data_text,  chunksize_text[i_chunk], 
		 1,i_chunk);

#ifdef STATS		
    clock_align += clock() - last_clock;
    last_clock =  clock();
#endif   	

  }/*end of  [4] (chunk loop) */
  
  free(last_tuple_pos_with_dist);
  free(last_tuple_ptr_with_dist);
  free(keylist_query);
  free(first_query);
      
  return 0;
}
  

