/*
 *  YASS 1.07
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>


#include "util.h"
#include "global_var.h"
#include "display.h"
#include "prdyn.h"
#include "tuple.h"
#include "proba.h"
#include "kword.h"

int stat1[3]      ={0,0,0};
int stat2[3]      ={0,0,0};
int transindels[3]={0,0,0};

int Display_Alignements(int querysize,
			char * dataquery,
			char * dataqueryrev,char * queryname,
			char * datatext,
			int nbchunks, char ** chunkname, int * chunksize, int * chunkstart)
{

  int  nb_displayed_alignments = 0;
  MA * ma = NULL;
 

  /* [0] header if needed */
  if (gp_display == 2){
    fprintf(OUTSTREAM,"# Fields: Query id, Subject id, %% identity, alignment length, mismatches, gap openings, q. start, q. end, s. start, s. end, e-value, bit score\n");
    /* I hope this line is not copyrighted (because this is not mine)  ;-) */
  }
  if (gp_display == 3){
    fprintf(OUTSTREAM,"# q. start,\tq. end,\ts. start,\ts. end,\tq. length,\ts.length,\tf/r \n");
    /* Well this one is mine ... */
  }
    				       

  /* [1] Effective display of each Memorized Alignment */
  for ( ma = gv_first_MA ; ma != NULL && nb_displayed_alignments < gp_nbmaxlines ; ma=ma->next){

    nb_displayed_alignments++;
    
#ifdef DEBUG	
    DisplayMA(ma);
#endif	  
   
    switch(gp_display){
    case 0:
    case 1:
      /*  (1.1) alignment position repeat */ 
      if (!(ma->reverse))     
	fprintf(OUTSTREAM,"*(%i-%i)(%i-%i)",
		BCOUNT + ma->left_pos_begin,
		BCOUNT + ma->left_pos_end,
		BCOUNT + ma->right_pos_begin,
		BCOUNT + ma->right_pos_end);
      else
	fprintf(OUTSTREAM,"*(%i-%i)(%i-%i)",
		BCOUNT + querysize- ma->left_pos_begin,
		BCOUNT + querysize- ma->left_pos_end,
		BCOUNT + ma->right_pos_begin,
		BCOUNT + ma->right_pos_end );
    
      /* (1.2) print details (size, evalue, etc ...) */
      fprintf(OUTSTREAM," Ev: %g s: %i/%i %c\n", 
	      Evalue(gp_k_blast,gp_lambda_blast,gp_querysize,gp_textsize,ma->blastscore),
	      ma->left_pos_end  - ma->left_pos_begin  + 1,
	      ma->right_pos_end - ma->right_pos_begin + 1,
	      (ma->reverse)?'r':'f'
	      );
      
      if (gp_display==0)
	break;
    
      /* (2.1) print chunk names (reverse/complement) */ 	
      fprintf(OUTSTREAM,"* %s %s / %s (forward strand)\n",
	      queryname,
	      (ma->reverse)?"(reverse complementary strand)":"(forward strand)",
	      chunkname[ma->i_chunk]
	      );
      /* (2.2) score bitscore and statistics */
      fprintf(OUTSTREAM,"* score = %i : bitscore = %.2f\n",(ma->blastscore),(float)BitScore(gp_k_blast,gp_lambda_blast,(ma->blastscore)));
      memset(stat1      ,'\0',3*sizeof(int));
      memset(stat2      ,'\0',3*sizeof(int));
      memset(transindels,'\0',3*sizeof(int));
      
      alignment_SG_stats_on_MA((ma->reverse)?dataqueryrev:dataquery
			       ,datatext + chunkstart[ma->i_chunk]
			       ,ma,transindels,stat1,stat2);
      
      fprintf(OUTSTREAM,"* mutations per triplet %i, %i, %i (%.2e) | ts : %i tv : %i\n\n",
	      stat1[0],stat1[1],stat1[2],
	      (float)(P_mutation_bias(stat1)),
	      transindels[0],transindels[1]
	      );
      /* (2.3) display alignment */
      display_alignment_SG_on_MA((ma->reverse)?dataqueryrev:dataquery
				 ,datatext + chunkstart[ma->i_chunk]
				 ,ma);     
      
      break;/* case 1: */
      
    case 2:
      /* (3) blast like tabular format */
      /* (3.1) fasta name */
      fprintf(OUTSTREAM,"%s\t%s\t", queryname,chunkname[ma->i_chunk]);
      /* (3.2) stats analysis */
      memset(stat1      ,'\0',3*sizeof(int));
      memset(stat2      ,'\0',3*sizeof(int));
      memset(transindels,'\0',3*sizeof(int));
      
      alignment_SG_stats_on_MA((ma->reverse)?dataqueryrev:dataquery
			       ,datatext + chunkstart[ma->i_chunk]
			       ,ma,transindels,stat1,stat2);
      /* (3.3) percent identity */
      fprintf(OUTSTREAM,"%3.2f\t", 100 - 100.000 * ((double)(transindels[0]+transindels[1]+transindels[2])/MAX((ma->left_pos_end  - ma->left_pos_begin  + 1),(ma->right_pos_end - ma->right_pos_begin + 1))));
      /* (3.4) size */
      fprintf(OUTSTREAM,"%i\t",MAX((ma->left_pos_end  - ma->left_pos_begin  + 1),(ma->right_pos_end - ma->right_pos_begin + 1)));
      /* (3.5) mutations indels */
      fprintf(OUTSTREAM,"%i\t%i\t",transindels[0]+transindels[1],transindels[2]);
      /* (3.6) alignment position repeat */ 
      if (!(ma->reverse))
	fprintf(OUTSTREAM,"%i\t%i\t%i\t%i\t",
		BCOUNT + ma->left_pos_begin,
		BCOUNT + ma->left_pos_end,
		BCOUNT + ma->right_pos_begin,
		BCOUNT + ma->right_pos_end);
      else
	fprintf(OUTSTREAM,"%i\t%i\t%i\t%i\t",
		BCOUNT + querysize- ma->left_pos_end,
		BCOUNT + querysize- ma->left_pos_begin,
		BCOUNT + ma->right_pos_end,  /* inverted with the next line */
		BCOUNT + ma->right_pos_begin);
      fprintf(OUTSTREAM,"%.2g\t%.3g\n", Evalue(gp_k_blast,gp_lambda_blast,gp_querysize,gp_textsize,ma->blastscore),BitScore(gp_k_blast,gp_lambda_blast,(ma->blastscore)));
      break;/* case 2: */
    case 3:
       /*  (1.1) alignment position repeat */ 
      if (!(ma->reverse))     
	fprintf(OUTSTREAM,"%i\t%i\t%i\t%i",
		BCOUNT + ma->left_pos_begin,
		BCOUNT + ma->left_pos_end,
		BCOUNT + ma->right_pos_begin,
		BCOUNT + ma->right_pos_end);
      else
	fprintf(OUTSTREAM,"%i\t%i\t%i\t%i",
		BCOUNT + querysize- ma->left_pos_begin,
		BCOUNT + querysize- ma->left_pos_end,
		BCOUNT + ma->right_pos_begin,
		BCOUNT + ma->right_pos_end );
    
      /* (1.2) print details (size, evalue, etc ...) */
      fprintf(OUTSTREAM,"\t%i\t%i\t%c\n", 
	      ma->left_pos_end  - ma->left_pos_begin  + 1,
	      ma->right_pos_end - ma->right_pos_begin + 1,
	      (ma->reverse)?'r':'f'
	      );
      break;
      
    }/* switch */
  }/* for */
  return 0;
}

	


int Display_Stats(){
  fprintf(stderr,"* Statistical parameters\n");
  fprintf(stderr,"   rho = %i\n",gp_rho_stat);
  fprintf(stderr,"   delta = %i\n",gp_delta_stat);
#ifdef STATS
  fprintf(stderr,"   ----\n");
  fprintf(stderr,"   nb single seeds        = %i\n",nb_seeds);
  fprintf(stderr,"   nb chains built        = %i\n",nb_chains_built);
  fprintf(stderr,"   nb alignments tested   = %i\n",nb_chains_tested);
  fprintf(stderr,"   nb alignments detected = %i\n",nb_ma);
  fprintf(stderr,"   ----\n");
  fprintf(stderr,"   nb grouping tests      = %i\n",nb_postprocessed_grouping_tests);
  fprintf(stderr,"   nb alignments grouped  = %i\n",nb_postprocessed_grouping_links);
  fprintf(stderr,"   nb alignments found    = %i\n",nb_postprocessed_ma);

  if (clock_align+clock_chain+clock_post == 0)
    clock_align=clock_chain=clock_post = 1;    
   
  fprintf(stderr,"   time spent = %i s\n",time_spent);
  fprintf(stderr,"       chaining  time = %i s [ %i%%]\n"
	  ,(int) ( 0.5 + time_spent * (clock_chain/(clock_align+clock_chain+clock_post)))
	  ,(int) ( 0.5 + 100        * (clock_chain/(clock_align+clock_chain+clock_post)))
	  );
  fprintf(stderr,"       alignment time = %i s [ %i%%]\n"
	  ,(int) ( 0.5 + time_spent * (clock_align/(clock_align+clock_chain+clock_post)))
	  ,(int) ( 0.5 + 100        * (clock_align/(clock_align+clock_chain+clock_post)))
	  );
  fprintf(stderr," post-processing time = %i s [ %i%%]\n"
	  ,(int) ( 0.5 + time_spent * (clock_post/(clock_align+clock_chain+clock_post)))
	  ,(int) ( 0.5 + 100        * (clock_post/(clock_align+clock_chain+clock_post)))
	  );
#endif	 
  return 0;
}




int Display_Params(){
  fprintf(stderr,"\n");
  fprintf(stderr,RULE);
  
  /* filenames */
  if (gp_nbfiles <= 1){
    fprintf(stderr,"DNA file : ");
    fprintf(stderr,"\"%s\"\n",gp_files[0]);
  }else{
    fprintf(stderr,"first DNA file : ");
    fprintf(stderr,"\"%s\"",gp_files[0]);
    fprintf(stderr,", second DNA file : ");
    fprintf(stderr,"\"%s\"\n",gp_files[1]); 
  }

  fprintf(stderr,"* Parameters\n");
  fprintf(stderr,"  seed pattern \"%s\" : weight = %i bits, span = %i bases\n",gp_motif,gp_seed_bitweight , gp_seed_span); 
  fprintf(stderr,"  mutation rate = %i%%\n",gp_mutations_percent);
  fprintf(stderr,"  indels rate = %i%%\n",gp_indels_percent);
  fprintf(stderr,"  tolerance  = %i%%\n",gp_alpha_percent);
  fprintf(stderr,"  minimum score seeds = %i\n",gp_minimum_score_seeds); 
  return 0;
}




/* Dot progress bar : 
 * "launch yass and lets take a coffee" progress bar
 */

int last_point    = 0;			 

int Display_Progress(unsigned int pos, unsigned int maxpos)
{  
  if (gp_reverse == 2){
    if (gv_reverse_repeat){
      if (last_point  *  maxpos < pos  *  (NBL/2) + maxpos * (NBL/2)){
	while(last_point  *  maxpos < pos  *  (NBL/2) + maxpos * (NBL/2) && last_point < NBL){
	  fprintf(stderr,".");fflush(NULL);
	  last_point++;
	}
      } 
    }else{
      if (last_point  *  maxpos < pos  *  (NBL/2)){
	while(last_point  *  maxpos < pos  *  (NBL/2) && last_point < NBL){
	  fprintf(stderr,".");fflush(NULL);
	  last_point++;
	}
      }
    }
  }else{
    if (last_point  *  maxpos < pos  *  (NBL)){
      while(last_point  *  maxpos < pos  *  (NBL) && last_point < NBL){
	fprintf(stderr,".");fflush(NULL);
	last_point++;
      }
    }
  }
  return 0;
}
  
  

