/*
 *  YASS 1.07
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifndef  __GLOBAl_VAR_H_
#define  __GLOBAl_VAR_H_
#include <stdio.h>
#include "util.h"
#include "tuple.h"

/********************
 * GLOBALVARIABLES  *
 ********************/


/* [1]  MODIFIED WHEN STARTING PROGRAM 
 *========================================*/

/* parameters and flags */
extern int   gp_seed_bitweight      ;
extern int   gp_seed_span           ;
extern char  gp_motif[]             ;
extern int * gp_automata            ;

extern int  gp_minimum_score_seeds  ;
extern int    gp_minscore           ;
extern int    gp_mutations_percent  ;
extern double gp_mutations          ;
extern int     gp_indels_percent    ;
extern double  gp_indels            ;
extern int    gp_alpha_percent      ;
extern double gp_alpha              ;

extern double gp_entropy_min        ;
extern double gp_expectation_value  ;

extern int  gp_reverse              ;
extern int  gp_display              ;
extern int  gp_nbmaxlines           ;
extern int  gp_distdiag             ;
extern int  gp_t                    ;

/* statisticals parameters */

extern int  gp_rho_stat             ;
extern int  gp_delta_stat           ;


/* post-processing */
extern int gp_win_min               ;
extern int gp_win_max               ;
extern int gp_win_mul               ;

/* files number et filenames */
extern int  gp_nbfiles              ;
extern char gp_files[2][256]        ;
extern int  gp_selection_fasta      ;


/* cost */
extern double      KBLASTTABLE[];
extern double LAMBDABLASTTABLE[];
extern int    SUBMATRIXTABLE[4][4];
extern int    INDELSTABLE   [4][2];
extern int  gp_cost_gap_opened     ;
extern int  gp_cost_gap_continued  ;
extern int  gp_cost_match          ;
extern int  gp_cost_mismatch       ;
extern int    gp_cost_transition   ;
extern int    gp_cost_other        ;
extern int  gp_substitution_matrix[16][16];
extern int  gp_cost_max_substitution_matrix;
extern double gp_k_blast          ;
extern double  gp_lambda_blast     ;
extern int     gp_xdrop;

/* data tables and their sizes  */
extern char *  gp_query;
extern char *  gp_text;
extern int gp_querysize;
extern int gp_textsize;
extern int gp_querysize_chunk;


/* [2]   MODIFIED WHEN RUNNING ...
 *================================*/
extern FILE *      gv_outstream;
extern int         gv_reverse_repeat; /* flag to register all currently found repeats as "reversed ones"  */
extern tuplelist * gv_last_tl ; /* last tuple list added by the "assemble" algorithm */


extern MA * gv_first_MA ;      /* MA in the current list */
extern MA * gv_last_MA  ;
extern int  gv_i_current;       /* used by "assemble" and read by "align" */

extern int  gv_left_correction; /* left correction factor (0 if assemble_single ... otherwise != 0) */

/* statistical variables */

/* statistics on proccessing step */
extern int nb_seeds       ;
extern int nb_chains_built;
extern int nb_chains_tested;
extern int nb_ma           ;

/* statistics on post processing step */
extern int nb_postprocessed_grouping_tests;
extern int nb_postprocessed_grouping_links;
extern int nb_postprocessed_ma;


extern int    time_spent           ; /* full time spent (seconds) */
extern double last_clock           ; /* last time measured */
extern double clock_align          ; /* time spent to alignement (cycles of 1/1000000 s) */
extern double clock_chain          ; /* time spent to chain      (cycles of 1/1000000 s) */
extern double clock_post           ;

extern char lookup[16] ;
#ifdef GORE_ALLOC_TUPLE
extern tuple *     tupler      ;  
extern int         i_tuple     ;
#endif	
#ifdef GORE_ALLOC_TUPLELIST
extern tuplelist * tuplelister ;
extern int         i_tuplelist ;
#endif

#endif
