/*
 *  YASS 1.07
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h> 
#include <string.h>      
#include <errno.h>
#include <math.h>
#include "util.h"
#include "kword.h"
#include "global_var.h"



/*
 *   Build a  DNA bases table  (code 0..3) from a DNA file
 * 
 * - Input  : "filename"                                    : DNA file name
 * - Output : "p_data"                                      : return the allocated data table pointer 
 *            "p_datasize"                                  : return the allocated data table size 
 *            "p_nbchunks","p_chunkname", and "p_chunksize" : return respectively the number of chunks fond in the multifasta file, their names and their size
 *                                                            (set to NULL if you do not need these informations
 */


int CreateData(/*in*/ char * filename,/*out*/ char ** p_data,/*out*/ int * p_datasize,/*out*/int * p_nbchunks,/*out*/char *** p_chunkname ,/*out*/ int ** p_chunksize,/*out*/ int ** p_chunkstart){
  int    i = 0,    j = 0 , firstspace = 0;
  int    car         = 0 ;
  int    datasize    = 0 ;
  int    cumulength  = 0 ;
  char * data        = NULL;
  FILE * file        = NULL;
  
  
  int     nbchunks   = 0 ;
  char ** chunkname  = NULL;
  int  *  chunksize  = NULL;
  int  *  chunkstart = NULL;
  
  /* (1) count number of DNA bases availaible */
  
/*
 *      A --> adenosine           M --> A C (amino)
 *      C --> cytidine            S --> G C (strong)
 *      G --> guanine             W --> A T (weak)
 *      T --> thymidine           B --> G T C
 *      U --> uridine             D --> G A T
 *      R --> G A (purine)        H --> A C T
 *      Y --> T C (pyrimidine)    V --> G C A
 *      K --> G T (keto)          N --> A G C T (any)
 */

  if(!(file = fopen(filename,"r"))){
    fprintf(stderr,"* unable to open \"%s\": errno = %i \n",filename,errno);
    exit(-1);
  }
   /* comment */
  while((car=fgetc(file)) != EOF ){
    if (car == EOF) break;
    switch(car){
    case 'A':
    case 'a':
    case 'T':
    case 't':
    case 'G':
    case 'g':
    case 'C':
    case 'c':
    case 'U':
    case 'u':  
    case 'R':
    case 'r':
    case 'Y':
    case 'y':       
    case 'K':
    case 'k':
    case 'M':
    case 'm':
    case 'S':
    case 's':
    case 'W':
    case 'w':
    case 'B':
    case 'b':
    case 'D':
    case 'd':
    case 'H':
    case 'h':
    case 'V':
    case 'v':
    case 'N':
    case 'n':
      datasize++;
      break;
    case ' ':
    case '\t':
    case '\n':
      break;

    case '>':
      nbchunks++;
      do{
	/* seek eol */
	car = fgetc(file);
      }while ( car != '\n' && car != EOF );
      break;
    default:
      while ((car=fgetc(file))!='\n');
      break;
    } /*switch*/
  }/*while*/
 

  if(!datasize){
    fprintf(stderr,"* \"%s\" is empty\n",filename);
    exit(-1);
  }

  fclose(file);
    
  /* (2) allocate one memory blok for the DNA data */
  
  data = (char *)calloc((unsigned)datasize,sizeof(char));
  ASSERT(data,CreateData);
  
  

  if (nbchunks > 0){
    chunkname =  (char **)calloc((unsigned)nbchunks,sizeof(char *));
    ASSERT(chunkname,CreateData);
    chunksize =  (int   *)calloc((unsigned)nbchunks,sizeof(int));
    ASSERT(chunksize,Createdata);  
    chunkstart =  (int   *)calloc((unsigned)nbchunks,sizeof(int));
    ASSERT(chunkstart,Createdata);
  }
  


  /* (3) fill the DNA data */  
  
  if(!(file = fopen(filename,"r")))
    return -3;
  /* description of sentance */
  while((car=fgetc(file)) != EOF ){
    if (car == EOF) break;
    switch(car){
    case 'A':
    case 'a':     
      data[i++] = (char) 0;
      break;
    case 'T':
    case 't':
       data[i++] = (char) 1;
      break;
    case 'G':
    case 'g':
      data[i++] = (char) 2;
      break;
    case 'C':
    case 'c':
      data[i++] = (char) 3;
      break;
    case 'Y':
    case 'y':     
      data[i++] = (char) 4;
      break;
    case 'S':
    case 's':
      data[i++] = (char) 5;
      break;
    case 'U':
    case 'u':
      data[i++] = (char) 6;
      break;
    case 'N':
    case 'n':      
      data[i++] = (char) 7;
      break;
    case 'R':
    case 'r': 
      data[i++] = (char) 8;
      break;
    case 'M':
    case 'm':
      data[i++] = (char) 9;
      break;
    case 'W':
    case 'w':
      data[i++] = (char)10;
      break;
    case 'D':
    case 'd':
      data[i++] = (char)11;
      break;
    case 'H':
    case 'h':
      data[i++] = (char)12;
      break;
    case 'V':
    case 'v':
      data[i++] = (char)13;
      break;
    case 'K':
    case 'k':
      data[i++] = (char)14;
      break;
    case 'B':
    case 'b':
      data[i++] = (char)15;
      break;
    case ' ':
    case '\t':
    case '\n':
      break; 
    case '>':
      /* --[A] chunk name / size update */
      if (j>0){
	chunksize[j-1] = i - cumulength;
      }
      cumulength     = i;
      chunkstart[j] = i;
      chunkname[j]  = (char *)malloc((unsigned)256*sizeof(char));
      fgets(chunkname[j],256,file); 
      (chunkname[j])[strlen(chunkname[j])-1]='\0';
      for(firstspace = 0 ; firstspace < 256 ; firstspace++){
	if ((chunkname[j])[firstspace] == ' ')
	  (chunkname[j])[firstspace] = '\0';
      }     
      j++;
      break;
    default:
      while ((car=fgetc(file))!='\n');
      break;
    } /*switch*/
  }/*while*/
  fclose(file);

  /* --[B] last chunk size updated */
  if (nbchunks > 0){
    if (j>0){
      chunksize[j-1] = i - cumulength;
    }
  }
  /* --[C] Raw format : simulate one big fasta chunk */
  if (nbchunks == 0){
    nbchunks = 1;
    chunkname =  (char **)calloc((unsigned)nbchunks,sizeof(char *));
    ASSERT(chunkname,CreateData);
    chunksize =  (int   *)calloc((unsigned)nbchunks,sizeof(int));
    ASSERT(chunksize,Createdata);
    chunkstart =  (int   *)calloc((unsigned)nbchunks,sizeof(int));
    ASSERT(chunkstart,Createdata);
    chunkname[0]  = filename;
    chunksize[0]  = datasize;
    chunkstart[0] = 0;
  }
  

#ifdef DEBUG
  for(i=0;i<nbchunks;i++)
    fprintf(stderr," chunksize = %i chunkstart = %i chunkname = \"%s\" \n",chunksize[i],chunkstart[i],chunkname[i]);
#endif	

  /* (4) return allocated and computed data*/
  *p_datasize  = datasize;
  *p_data      = data;
  if (p_nbchunks != NULL)
    *p_nbchunks  = nbchunks;
  if (p_chunkname != NULL)
    *p_chunkname = chunkname;
  if (p_chunksize != NULL)
    *p_chunksize = chunksize; 
  if (p_chunkstart != NULL)
    *p_chunkstart = chunkstart;
  return 0;
}



/*
 *   Build a DNA bases of the complementry reverse straind 
 * 
 * - Input  : "data_in"     : DNA array (returned by CreateData for example)
 *            "datasize"    : DNA array size
 * - Output : "p_data_out"  : DNA array containing the complementary reverse straind
 *            
 */


int CreateReverseComplement(/*in*/ char * data_in ,/*in*/ int  datasize, /*out*/ char ** p_data_out){  
  int i;
  char *  data = (char *)calloc((unsigned)datasize,sizeof(char));
  ASSERT(data,CreateReverseComplement);
  
  for(i=0;i<datasize;i++)
    data[datasize-1-i] = COMPLEMENT(data_in[i]);
  
  *p_data_out     = data;
  return 0;
}


int KeyComplexity(int key)
{
  return 1;
}


/*
 * Pattern motif size computation (gp_seed_bitweight, gp_seed_span)
 *
 */


int ComputeLength()
{
  int i;
  gp_seed_bitweight  = 0;
  gp_seed_span       = 0;
  
  for(i=0;i<(int)strlen(gp_motif);i++){
    if (gp_motif[i]=='#')
      (gp_seed_bitweight)+=2;
    if (gp_motif[i]=='@')
      (gp_seed_bitweight)+=1;
    (gp_seed_span)++;
  }
  gp_motif[gp_seed_span] = '\0';
  return 0;
}

void MotifAdd(char * res,char * part)
{
  char * p = part;
  char * r = res;

  while(*p){ 
    switch(*p){
     
    case '#':
      *r='#';
      break;
      
    case '@':
      if (*r!='#')
	*r='@';
      break; 
    } 
    p++;
    r++;
  }
}

int MotifSub(char * buffer)
{
  int i = 0;
  int result = 0;
  
  switch(gp_motif[gp_seed_span-1]){
  case '#':
    result = 2;
    break;
  case '@':
    result = 1;
    break;
  }

  for(i = gp_seed_span - 1; i > 0; i--){
    switch (buffer[i]){
    case '#':
      break;
    case '@':
      if (gp_motif[i-1]=='#')
	result += 1;
      break;
    case '_':
      switch(gp_motif[i-1]){
      case '#':
	result += 2;
	break;
      case '@':
	result += 1;
	break;
      }
      break;
    }
  }
  return result;
}

   

void  AutomataOverlaps(int delph)
{
  int i,j;
  int p_delph      = 1<<delph;
  char * buffer    = (char * )malloc(gp_seed_span+1); 
  
  ASSERT(buffer,AutomataOverlaps);
  memset(buffer,'_',gp_seed_span);
  buffer[gp_seed_span]='\0';
  
  gp_automata = (int  * )malloc(sizeof(int)*p_delph);
  ASSERT(gp_automata,AutomataOverlaps);
   
  /* for each possible configuration */
  for(i=0;i<p_delph;i++){
   
    /* overlapped elements count */ 
    for(j=0;j<delph;j++){
      if(i&(1<<j)){
	 MotifAdd(buffer,gp_motif+j);
      }
    }
    gp_automata[i] = MotifSub(buffer);
  }
  free(buffer);
}
   

int Key(/*in*/char * data, /*in*/ int pos)
{
  int i = 0;
  int keyvalue  = 0;
  int puissance = 0;
  for(i=0;i<gp_seed_span;i++){
    if ((int)data[pos+i]>3)
      return 0;
    if (gp_motif[i] == '#'){
      keyvalue  |= ((int)data[pos+i]) << puissance;
      puissance += 2;
    }
    if (gp_motif[i] == '@'){
      keyvalue  |= ((int)(data[pos+i])&0x01) << puissance; /* purine/pyrymidine */
      puissance += 1;
    }
  }
  if (keyvalue>16383) {
    int keyvalue  = 0;
    int puissance = 0;
    for(i=0;i<gp_seed_span;i++){
      if ((int)data[pos+i]>4)
	return 0;
      if (gp_motif[i] == '#'){
	keyvalue  |= ((int)data[pos+i]) << puissance;
      puissance += 2;
      }
      if (gp_motif[i] == '@'){
	keyvalue  |= ((int)(data[pos+i])&0x01) << puissance; /* purine/pyrymidine */
	puissance += 1;
      }
    }
  }
  return keyvalue;
}



/*
 * Create a linkedlist of kwords with param "list" and index "first"
 *
 * - Entre: "data"    : dna input table 
 *           "datasize": dna input table size
 * - Sortie: "p_first" : return a table of size (2^w). For each code w, it contains the first 
 *                       position occurence of the seed with such code on the sequence
 *           "p_list"  : return a table of size (keysize). For each position i, it contains the next
 *                       position j such that seed i and j code are equal. 
 */


int CreateKeyList(/*in*/ char * data,/*in*/ int datasize, /*out*/ int ** p_first, /*out*/ int ** p_list){
    
  int i,j,w;
  int *bucklet = NULL;
  int *first   = NULL;
  int *list    = NULL;
  int *listSort=NULL;
  int *count   = NULL;
  int sizeofbucklet = 1;
  int keysize = datasize - gp_seed_span + 1;
  int maxchainsize = 0;
  
  

  /* (1) allocation of 3 table */
  sizeofbucklet <<= (gp_seed_bitweight);

  bucklet = (int *)calloc((unsigned)sizeofbucklet,sizeof(int));
  ASSERT(bucklet,CreateKeyList);
  memset(bucklet,0xff,(unsigned)sizeofbucklet*sizeof(int));
  
  
  first  =  (int *)malloc((unsigned)(sizeofbucklet+1)*sizeof(int));
  ASSERT(first,CreateKeyList);
  memset(first,0xff,(unsigned)(sizeofbucklet+1)*sizeof(int));

  list   =  (int *)calloc((unsigned)keysize+1,sizeof(int));
  ASSERT(list,CreateKeyList);
  memset(list,0xff,(unsigned)(keysize+1)*sizeof(int));
  
  count  =  (int *)calloc((unsigned)sizeofbucklet,sizeof(int));
  ASSERT(count,CreateKeyList);
  memset(count,0x0,(unsigned)sizeofbucklet*sizeof(int));

  listSort =  (int *)malloc((unsigned)keysize*sizeof(int));
  ASSERT(listSort,CreateKeyList);
  memset(listSort,0xff,(unsigned)keysize*sizeof(int));




  /* (1.1) suppression d'elements issus de la keycomplexity */
  /*
  for(w=0;w<sizeofbucklet;w++){
   
    if (!KeyComplexity(w))
      bucklet[w] = -2;
  }
  */
  /* (2) list creation alogorithm */
  for(j=0;j<keysize;j++){
    w  = Key(data,j); 
    if (bucklet[w] >= -1){  
      count[w]++; 
      maxchainsize =  MAX(count[w],maxchainsize);
      if( (i=bucklet[w]) == -1)
	first[w] = j;
      else	
	list[i] = j;
      bucklet[w] = j;
    }
  }
  free(bucklet);
  
  /** Truncate highly repetitive case*/
  first[0              ]              = -1; /* all "A" code */
  first[sizeofbucklet-1]              = -1; /* all "C" code */
  for( w=0 ; w < sizeofbucklet; w++){ 
    if( count[w] > MAX(maxchainsize*(100-gp_t)/100,2000) ){
      first[w] = -1;
    }
  }
  free(count);


  /** Pool a index for a as code 
  **/
  i=0;
  for( w=0 ; w <sizeofbucklet ; w++){
    j=first[w];
    first[w]=i;
    if (j!=-1) {
      listSort[i]=j;
      i++;
      while ( (j=list[j])!=-1){
	listSort[i]=j;
	i++;
      }
    }
  }
  first[w]=i;
  free(list);
  *p_list  = listSort;
  *p_first = first   ;
  return 0;  
}



/*
 * Display sequence fragments
 */

int DisplaySequence(char * data, int pos, int len)
{
  int j;

  for(j=pos;j<pos+len;j++){
    printf("%c",lookup[(int)data[j]]);
  } 
  printf("\n");
  return 0;
}

/*
 * COMPLEXITY CHECK
 */

double Entropy(char * data,int pos,int len,int k)
{
  int p4k  = 1<<(k<<1); /*4^k*/
  int mask = 0;
  int key  = 0;
  int i    = 0;
  int * freq = (int *)malloc(p4k*sizeof(int));
  double result = 0.00;
  ASSERT(freq,Entropy);
    
  for(i=0;i<p4k;i++)
    freq[i] = 0;

  for(i=0;i<k;i++){
    mask <<= 2;
    mask  |= 3;
    key  <<= 2;
    key   |= (int)data[pos+i];
  }
  
  
  for(i=k;i<len;i++){
    key <<= 2;
    key  |= (int)data[pos+i];
    key  &= mask;
    freq[key]++;
  }

  
 
  for(i=0;i<p4k;i++){
    if (freq[i] != 0){
      result -= ((double)(freq[i])/(double)(len-k+1)) * log((double)(freq[i])/(double)(len-k+1));
    }
  }
  result /= log(2.0000);
  free(freq);
  return result;
}

