/*
 *  YASS 1.07
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "global_var.h"
#include "tuple.h"
#include "prdyn.h"

#define  CHAR_EQU '|'
#define  CHAR_SS  '.'
#define  CHAR_SV  ':'
#define  CHAR_DEL ' '
#define  CHAR_INS ' '
#define  CHAR_DUM '-'

#define GAP1 1  /* gap cost */
#define SUB1 2  /* substitution cost */

int * buffer01 = NULL ;


/*
 *  [I]
 *
 *   "quick" alignement Functions used in the alignment part
 *
 ***********************************************************************/



/*
 * Run this method before any use of "quick" alignment algorithm
 * (it allocate dynamic programming table)
 *
 */

#ifdef INLINE
 inline 
#endif
int initialise_alignment(int size)
{
  buffer01 = (int *)malloc((unsigned)(size+1)* sizeof(int));
  ASSERT(buffer01,initialise_aligment);
#ifdef GORE_ALLOC_TUPLE
  tupler      = (tuple *) malloc((unsigned)(GORE_ALLOC_TUPLE_SIZE)* sizeof(tuple));
  ASSERT(tupler,initialise_aligment);
  memset(tupler,0x00,            (unsigned)(GORE_ALLOC_TUPLE_SIZE)* sizeof(tuple));
#endif
#ifdef GORE_ALLOC_TUPLELIST
  tuplelister = (tuplelist *) malloc((unsigned)(GORE_ALLOC_TUPLELIST_SIZE)* sizeof(tuplelist));
  ASSERT(tuplelister,initialise_aligment);
  memset(tuplelister,0x00,           (unsigned)(GORE_ALLOC_TUPLELIST_SIZE)* sizeof(tuplelist));
#endif
  return 0;
}


/*
 * Align two segments defined by their started pos and their length 
 */

#ifdef INLINE
 inline 
#endif
int alignment(char * data1, int pos1, int len1,char * data2, int pos2, int len2){
  
  int i,j;
  int tmp_diag,tmp_diag_prev;
  
  for(i=0;i<len1+1; i++)
    buffer01[i] = i*GAP1;
  
  /* columns */
  for(j=1;j<len2+1;j++){
    
    tmp_diag_prev = (j-1)*GAP1;
    buffer01[0] = j*GAP1;
    
    
    /* lines*/
    for(i=1;i<len1+1;i++){
      tmp_diag = buffer01[i];

      /* compute value */
      if (data2[pos2 + j-1] == 
	  data1[pos1 + i-1])
	buffer01[i] = MIN3(tmp_diag_prev,     buffer01[i]+GAP1, buffer01[i-1]+GAP1);
      else
	buffer01[i] = MIN3(tmp_diag_prev + SUB1, buffer01[i]+GAP1, buffer01[i-1]+GAP1);
      
      /* update diag value */
      tmp_diag_prev = tmp_diag;
    }
  } 
  return buffer01[len1];
}

/*
 * Try to align two subparts;
 * return -1 si failed ( "leavethreashold" is over "steps_to_leave") or 
 * the result (then /out/ positions are filled)
 */

#ifdef INLINE
 inline 
#endif
int left_alignment(char * data1, int pos1, int len1,
		   char * data2, int pos2, int len2, 
		   double leavethreshold,int steps_to_leave,
		   /*out*/ int * p_gain_score,/*out*/ int * p_last_pos1,/*out*/ int * p_last_pos2)
{
  int i,j;
  int tmp_diag,tmp_diag_prev;
  int steps_over_threshold = 0;
  int min_score_in_column = 0;
  int i_min = 0;
  

  for(i=0;i<len1+1; i++)
    buffer01[i] = i*GAP1;
  
  /* columns */
  for(j=1;j<len2+1;j++){
    
    tmp_diag_prev = (j-1)*GAP1;
    buffer01[0] = j*GAP1;
    
    min_score_in_column = INFINITY;
    
    
    /* lines */
    for(i=1;i<len1+1;i++){
      tmp_diag = buffer01[i];

      /* compute value */
      if (data2[pos2 + j-1] == data1[pos1 + i-1])
	buffer01[i] = MIN3(tmp_diag_prev,     buffer01[i]+GAP1, buffer01[i-1]+GAP1);
      else
	buffer01[i] = MIN3(tmp_diag_prev + SUB1, buffer01[i]+GAP1, buffer01[i-1]+GAP1);
       
      if (buffer01[i] < min_score_in_column){ 
	min_score_in_column = buffer01[i]; 
	i_min = i;
      }
      
      /* update diag value */
      tmp_diag_prev = tmp_diag;
    }

    if (((double)(min_score_in_column))/((double)j) > leavethreshold )
      steps_over_threshold++;
    if (steps_over_threshold > steps_to_leave){
      *(p_last_pos1) = pos1 - i_min;
      *(p_last_pos2) = pos2 - j;
      *(p_gain_score) = min_score_in_column;
      return -1;  
    }
  }/* for */

  *(p_last_pos1) = pos1 - i_min;
  *(p_last_pos2) = pos2 - j;
  *(p_gain_score) = min_score_in_column;

  return buffer01[len1];
}



/*
 * Try to align two subparts;
 * return -1 si failed ( "leavethreashold" is over "steps_to_leave") or 
 * the result (then /out/ positions are filled)
 */

#ifdef INLINE
 inline 
#endif
 int right_alignment(char * data1, int pos1, int len1,
			   char * data2, int pos2, int len2, 
			   double leavethreshold,int steps_to_leave,
			   /*out*/ int * p_gain_score,/*out*/ int * p_last_pos1,/*out*/ int * p_last_pos2)
{
  
  int i,j;
  int tmp_diag,tmp_diag_prev;
  int steps_over_threshold = 0;
  int min_score_in_column = 0;
  int i_min = 0;
  

  for(i=0;i<len1+1; i++)
    buffer01[i] = i*GAP1;
  
  /* columns */
  for(j=1;j<len2+1;j++){
    
    tmp_diag_prev = (j-1)*GAP1;
    buffer01[0] = j*GAP1;
    
    min_score_in_column = INFINITY;
    
    
    /* lines */
    for(i=1;i<len1+1;i++){
      tmp_diag = buffer01[i];

      /* compute value */
      if (data2[pos2 - j+1] == data1[pos1 - i+1])
	buffer01[i] = MIN3(tmp_diag_prev,     buffer01[i]+GAP1, buffer01[i-1]+GAP1);
      else
	buffer01[i] = MIN3(tmp_diag_prev + SUB1, buffer01[i]+GAP1, buffer01[i-1]+GAP1);
      
      if (buffer01[i] < min_score_in_column){ 
	min_score_in_column = buffer01[i]; 
	i_min = i;
      }
     
      /* update diag value */
      tmp_diag_prev = tmp_diag;
    }
    
    if (((double)(min_score_in_column))/((double)j) > leavethreshold )
      steps_over_threshold++;
    if (steps_over_threshold > steps_to_leave){
      *(p_last_pos1) = pos1 + i_min;
      *(p_last_pos2) = pos2 + j;
      *(p_gain_score) = min_score_in_column;
      return -1;  
    }
  }/* for */

  *(p_last_pos1) = pos1 + i_min;
  *(p_last_pos2) = pos2 + j;
  *(p_gain_score) = min_score_in_column;

  return buffer01[len1];
}




#define I2(i,j) (buffer01[(i)+ (((j)&1)?(len1+1):(0))])
#define M2(i,j) (buffer01[(i)+ (((j)&1)?(len1+1):(0)) + ((len1+1)*2)])
#define S2(i,j) (gp_substitution_matrix[ (int)(data1[pos1+(i)-1])][ ((int)(data2[pos2+(j)-1]))])

/*
 * Compute alignment score with :
 * - gap cost 
 * - gap extend
 * - subs 4x4 
 */

#ifdef INLINE
 inline 
#endif
 int alignment_SG(char * data1, int pos1, int len1,
		  char * data2, int pos2, int len2)
{
  int i,j;
  
  M2(0,0) = 0;
  I2(0,0) = gp_cost_gap_opened -  gp_cost_gap_continued;
  
  for(i=1;i<len1+1; i++)
    M2(i,0) = INFINITY;
  
  for(i=1 ; i<len1+1 ; i++)
    I2(i,0) = I2(i-1,0) +  gp_cost_gap_continued;
  
  for(j=1 ; j<len2+1 ; j++){
    
    M2(0,j) =  INFINITY;
    I2(0,j) =  I2(0,j-1) + gp_cost_gap_continued;
    
    for(i=1;i<len1+1;i++){
      I2(i,j) = MIN (
		     MIN( M2(i,j-1) + gp_cost_gap_opened    , M2(i-1,j) + gp_cost_gap_opened )
		     ,
		     MIN( I2(i,j-1) + gp_cost_gap_continued ,  I2(i-1,j) + gp_cost_gap_continued )
		     );
      M2(i,j) = MIN ( M2(i-1,j-1) + S2(i,j), I2(i-1,j-1) + S2(i,j) );
    }
  }
  return MIN(M2(len1,len2),I2(len1,len2));
}


#ifdef INLINE
 inline 
#endif
 int alignment_SG_Strait(char * data1, int pos1, char * data2, int pos2, int len)
{
  int i,result = 0;
  for(i=0; i < len ;i++){
    result += gp_substitution_matrix[(int)(data2[pos2+i])][(int)data1[pos1+i]];
  }
  return result;
}



#ifdef INLINE
 inline 
#endif
 int alignment_SG_on_MA(char * data1,  char * data2,
			MA * ma)
{
  int i,result = 0;
  int left_size,right_size;
  tuple * t = ma->first_tuple, * t_prev = ma->first_tuple ;
  
  /* [1] compute score before first tuple */
  
  left_size  = TBL_POS(t) - ma->left_pos_begin  + 1 ; /*  + 0 */
  right_size = TBR_POS(t) - ma->right_pos_begin + 1 ; /*  + 0 */
  
  if ( left_size > 0 && right_size > 0)
    result += alignment_SG(data1, ma->left_pos_begin , left_size ,
			   data2, ma->right_pos_begin, right_size
			   );
  
  /* [2] compute score for several tuples */
  while(t != NULL){
    /* (1) compute score inside tuple */
    for(i = 0 ;i < TSIZE(t) ; i++)
      result += gp_substitution_matrix[(int)(data1[i+TBL_POS(t)])][(int)(data2[i+TBR_POS(t)])];
	
    if (t->next){
      /* (2) comptue score between tuple (t) and (t->next) */
      result += alignment_SG(data1,TEL_POS(t),TGAP_L(t,t->next),
			     data2,TER_POS(t),TGAP_R(t,t->next));
    }/* if */
    t_prev = t;
    t = t->next;
  }/* while */

  /* [3] compute score after last tuple */
  
  left_size  =  ma->left_pos_end  - TEL_POS(t_prev) + 1;
  right_size =  ma->right_pos_end - TER_POS(t_prev) + 1;


  if ( left_size > 0 && right_size> 0)
     result += alignment_SG(data1, TEL_POS(t_prev), left_size ,
			    data2, TER_POS(t_prev), right_size 
			    );
  
  return result;
}


#ifdef INLINE
 inline 
#endif
 int alignment_SG_DROP(char * data1, int pos1, int len1,
			     char * data2, int pos2, int len2,
			     int XDROP )
{
  int i,j;
  int min;
  
  M2(0,0) = 0;
  I2(0,0) = gp_cost_gap_opened -  gp_cost_gap_continued;
  
  for(i=1;i<len1+1; i++)
    M2(i,0) = INFINITY;
  
  for(i=1 ; i<len1+1 ; i++)
    I2(i,0) = I2(i-1,0) +  gp_cost_gap_continued;
  
  for(j=1 ; j<len2+1 ; j++){
    
    M2(0,j) =  INFINITY;
    I2(0,j) =  I2(0,j-1)  + gp_cost_gap_continued;
    min     =  M2(0,j);
    
    for(i=1;i<len1+1;i++){
      I2(i,j) = MIN (
		     MIN( M2(i,j-1) + gp_cost_gap_opened    , M2(i-1,j) + gp_cost_gap_opened )
		     ,
		     MIN( I2(i,j-1) + gp_cost_gap_continued ,  I2(i-1,j) + gp_cost_gap_continued )
		     );
      M2(i,j) = MIN ( M2(i-1,j-1) + S2(i,j), I2(i-1,j-1) + S2(i,j) ); 
      min = MIN(min,M2(i,j));
    }
    if (min > XDROP)
      return min;
  }
  return MIN(M2(len1,len2),I2(len1,len2));
}



#define M2opt_current(a)   (*(buffer01 + d_current + (a)))
#define M2opt_previous(a)  (*(buffer01 + d_previous + (a)))
#define I2opt_current(a)   (*(buffer01 + I2_DELTA + d_current + (a))) 
#define I2opt_previous(a)  (*(buffer01 + I2_DELTA + d_previous + (a)))
#define S2opt(i,j) (gp_substitution_matrix[(int)(data1[pos1+(i)-1])][(int)(data2[pos2+(j)-1])])

#ifdef INLINE
 inline 
#endif
 int alignment_SG_DROP_opt(char * data1, int pos1, int len1,
			   char * data2, int pos2, int len2,
			   int XDROP )
{
  int i,j,s;
  int min;
  int I2_DELTA = (len1 + 1)<<1;
  int DELTA    = (len1 + 1);
  int d_previous = DELTA;
  int d_current  = 0;

  M2opt_current(0) = 0;
  I2opt_current(0) = gp_cost_gap_opened -  gp_cost_gap_continued;
  
  for(i=1;i<len1+1; i++){    
    M2opt_current(i) = INFINITY;
    I2opt_current(i) = I2opt_current(i-1) +  gp_cost_gap_continued;
  }
  

  for(j=1 ; j<len2+1 ; j++){
    if (j&1){
      d_previous = 0;
      d_current  = DELTA;
    }else{
      d_previous = DELTA;
      d_current  = 0;
    }
    
    
    M2opt_current(0) =  INFINITY;
    I2opt_current(0) =  I2opt_previous(0)  + gp_cost_gap_continued;
    min              =  INFINITY;
    
    for(i=1;i<len1+1;i++){ 
      d_previous ++;
      d_current  ++;
      I2opt_current(0) = MIN (
			      MIN( M2opt_previous(0) + gp_cost_gap_opened    , M2opt_current(-1) + gp_cost_gap_opened )
			      ,
			      MIN( I2opt_previous(0) + gp_cost_gap_continued , I2opt_current(-1) + gp_cost_gap_continued )
			      );
      s = S2opt(i,j);
      M2opt_current(0) = MIN ( M2opt_previous(-1) + s, I2opt_previous(-1) + s ); 
      min = MIN(min,M2opt_current(0));
    }
    if (min > XDROP)
      return min;
  }
  return MIN(M2opt_current(0),I2opt_current(0));
}





#define D_I2(k,i) (buffer01[(i) + ((k)&1?(MIN(len1,len2)+1):0) ])
#define D_M2(k,i) (buffer01[(i) + ((k)&1?(MIN(len1,len2)+1):0) + 2*(MIN(len1,len2)+1) ])
#define D_S2(i,j) (gp_substitution_matrix[(int)(data1[pos1+(i)])][(int)(data2[pos2+(j)])])


#ifdef INLINE
 inline 
#endif
 int  alignment_SG_Lz(char * data1, int pos1, int len1, char * data2, int pos2, int len2, int XDROP)
{
  
  int last_score  = 0, score  = 0;
  int last_indels = 0, indels = 0;
  int last_i      = 0, i      = 0;

  if (len1 < len2){
    
    for(i=0 ; i < len1 ; i++){
      score += D_S2(i,i+indels);
      if (score < last_score){
	last_score  = score;
	last_i      = i;
	last_indels = indels;
      }
      
      if (score > XDROP){
	score  = last_score + gp_cost_gap_opened + ABS(indels-last_indels) * gp_cost_gap_continued;
	i      = last_i;
	indels++;
	if (indels > len2 - len1)
	  return last_score;
      }
    }/* for i */
    return last_score;
  
  } else {
     
    for(i=0 ; i < len2 ; i++){
      score += D_S2(i+indels,i);
      if (score < last_score){
	last_score  = score;
       	last_i      = i;
	last_indels = indels;
      }
      
      if (score > XDROP){
	score  = last_score + gp_cost_gap_opened + ABS(indels-last_indels) * gp_cost_gap_continued;
	i      = last_i;
	indels++;
	if (indels > len1 - len2)
	  return last_score;
      }
    }/* for i */
    return last_score;
  }
  
}

/*
 * band alignment
 *
 *
 *          ------------------
 *   	    |^               |
 *          | bandwidth      |
 *          |v               |
 *          |----------------|	
 *          |I dlen          |
 *          |----------------|
 *          |^               |
 *          | bandwidth      |
 *          |v               |
 *	    ------------------
 *
 *
 *
 *  Classical DP matrix :
 *       (i) ->
 *     1  2  4  7  11 16 22
 * (j) 3  5  8  12 17 23
 *  |  6  9  13 18 24
 *  v  10 14 19 25
 *     15 20 26
 *     21 27
 *     28
 *
 *
 *
 *  Border DP matrix
 *  
 *               (k)->
 *     22
 *     16----------        (i,j-1)
 *(i)  11 23                     \
 * |   7  17            (i-1,j-1)-[*]
 * V   4  12 24                    |                   (i,j-1)
 *     2  8  18                  (i-1,j)                    \
 *---- 1  5  13  25 ----------------------------- (i-1,j-1)-[*]
 *     3  9  19                  (i,j-1)                    /
 *     6  14 26                    |                   (i-1,j)
 *     10 20            (i-1,j-1)-[*]
 *     15 27                     /
 *     21----------        (i-1,j)
 *     28
 */


int alignment_SG_Border(char * data1, int pos1, int len1, 
			char * data2, int pos2, int len2,
			int XDROP)
{
  /* band alignment length anw width*/
  int dwidth       = ABS(len2-len1);
  int bandwidth    = gp_delta_stat;
  
 
  /* various switch elements to access information */
  int row1       = 2*bandwidth + dwidth;
  int row2       = 2*row1;
  
  /* dp tables */
  int * bm0 = (int *)buffer01;
  int * bm1 = (int *)buffer01+row1;
  int * bi0 = (int *)buffer01+row2;
  int * bi1 = (int *)buffer01+row2+row1;
  
  /* maximal and minimal positions taken into account */
  int j,k;
  int jmin = 0, jmax = row1;
  
  for(j=0;j<row1;j++)
    bm0[j] = bi0[j] = (gp_cost_gap_opened) + ABS(j-bandwidth)*gp_cost_gap_continued;
  
  bm0[bandwidth] = bi0[bandwidth] = 0;

  if (len1 < len2){
    
    for(k=1 ; k< len1; k++){ 
     
      if(k&1){	
	/* M(j) full vectorisable loop */
	for(j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)(data2 + pos2)[j+k-bandwidth]][(int)((data1 + pos1)[k])];
	  bm1[j] = MIN( bm0[j] , bi0[j] ) + d; 
	}	
	
	/* MinxMax evaluations */
	j=jmin;	
	while( bm1[j++] > XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm1[j--] > XDROP && j>0)
	  jmax--;
	if (jmax<jmin)
	  return XDROP+1;
  

	/* I(j) in two half loops */
	bi1[bandwidth] = MIN(MIN(bi0[bandwidth-1],bi0[bandwidth+1]) + gp_cost_gap_continued , MIN(bm0[bandwidth-1],bm0[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi1[j-1] = MIN(
			 MIN(bi0[j-1],bi1[j]) + gp_cost_gap_continued,
			 MIN(bm0[j-1],bm1[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi1[j] = MIN(
		       MIN(bi0[j],bi1[j-1]) + gp_cost_gap_continued,
		       MIN(bm0[j],bm1[j-1]) + gp_cost_gap_opened
		       );
	}

      }else{/* k&1*/
          
      	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)(data2 + pos2)[j+k-bandwidth]][(int)((data1 + pos1)[k])];
	  bm0[j] = MIN( bm1[j] , bi1[j] ) + d;
	}	

	/* MinxMax evaluations */
	j=jmin;	
	while( bm0[j++] > XDROP && j<jmax)
	  jmin++;
	j=jmax;	
	while( bm0[j--] > XDROP && j>0)
	  jmax--;
	if (jmax<jmin)
	  return XDROP+1;
	

	/* I(j) in two half loops */
	bi0[bandwidth] = MIN(MIN(bi1[bandwidth-1],bi1[bandwidth+1]) + gp_cost_gap_continued , MIN(bm1[bandwidth-1],bm1[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi0[j-1] = MIN(
			 MIN(bi1[j-1],bi0[j]) + gp_cost_gap_continued,
			 MIN(bm1[j-1],bm0[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi0[j] = MIN(
		       MIN(bi1[j],bi0[j-1]) + gp_cost_gap_continued,
		       MIN(bm1[j],bm0[j-1]) + gp_cost_gap_opened
		       ); 
	}
      }/* k&1*/
    }/* k */

  }else{
    
    for( k=1 ; k< len2; k++){ 
          
      if(k&1){	
	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data2 + pos2)[k])][(int)(data1 + pos1)[j+k-bandwidth]];
	  bm1[j] = MIN( bm0[j] , bi0[j] ) + d;
	}	

	/* MinxMax evaluations */
	 j=jmin;	
	while( bm1[j++] > XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm1[j--] > XDROP && j>0)
	  jmax--;
	
	if (jmax<jmin)
	  return XDROP+1;

	/* I(j) in two half loops (each on is vectorisable on "non stupid" compilers) */
	bi1[bandwidth] = MIN(MIN(bi0[bandwidth-1],bi0[bandwidth+1]) + gp_cost_gap_continued , MIN(bm0[bandwidth-1],bm0[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi1[j-1] = MIN(
			 MIN(bi0[j-1],bi1[j]) + gp_cost_gap_continued,
			 MIN(bm0[j-1],bm1[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi1[j] = MIN(
		       MIN(bi0[j],bi1[j-1]) + gp_cost_gap_continued,
		       MIN(bm0[j],bm1[j-1]) + gp_cost_gap_opened
		       );
	}

      }else{/* k&1*/
          
      	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data2 + pos2)[k])][(int)(data1 + pos1)[j+k-bandwidth]];
	  bm0[j] = 
	    MIN( bm1[j] , 
		 bi1[j] ) 
	    + d;
	}	

	/* MinxMax evaluations */
	j=jmin;
	while( bm0[j++] > XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm0[j--] > XDROP && j>0)
	  jmax--;
	
	if (jmax<jmin)
	  return XDROP+1;
	
	/* I(j) in two half loops (each on is vectorisable on "non stupid" compilers) */
	bi0[bandwidth] = MIN(MIN(bi1[bandwidth-1],bi1[bandwidth+1]) + gp_cost_gap_continued , MIN(bm1[bandwidth-1],bm1[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi0[j-1] = MIN(
			 MIN(bi1[j-1],bi0[j]) + gp_cost_gap_continued,
			 MIN(bm1[j-1],bm0[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi0[j] = MIN(
		       MIN(bi1[j],bi0[j-1]) + gp_cost_gap_continued,
		       MIN(bm1[j],bm0[j-1]) + gp_cost_gap_opened
		       ); 
	}
      }/* k&1*/
    }/* k */
 
  }
  return (MAX(len1,len2)&1)?(bm0[bandwidth + dwidth]):(bm1[bandwidth + dwidth]);
}





  






/*
 * left/rigth extension if score does not fall down under XDROP
 * 
 */

#ifdef INLINE
 inline 
#endif
 int left_alignment_SG (char * data1, int pos1, int len1,
			char * data2, int pos2, int len2, 
			int XDROP, 
			/*out*/ int * p_gainscore,/*out*/ int * p_last_pos1,/*out*/ int * p_last_pos2)
{
  int i,j;
  int xdrop_current,xdrop_prev,xdrop_current2,xdrop_best=INFINITY;
  int d_i_prev,d_i;
  int m_i_prev,m_i;

  /* (1) initialisation */
  for(i=0;i<MIN(len1,len2);i++){
    D_I2(0,i) = INFINITY;
    D_I2(1,i) = INFINITY;
    D_M2(0,i) = INFINITY;
    D_M2(1,i) = INFINITY;
  }
 
  D_M2(0,0) = 0;   
  *(p_gainscore) = 0;
  *(p_last_pos1) = pos1;
  *(p_last_pos2) = pos2;
  
  xdrop_current2 = 0;
  xdrop_current  = INFINITY;
  xdrop_best     = 0;
  
  /* (2) compute (stop if Xdrop) */
  for(j=0;j<MIN(len1,len2);j++){/* for each diagonal */
    
    d_i = INFINITY;
    m_i = INFINITY;
    
    xdrop_prev    = xdrop_current;
    xdrop_current = INFINITY; 
    
    for(i=0;i<=j+1;i++){
       
      d_i_prev = d_i;
      m_i_prev = m_i;
      
      if (i<j){
	d_i      = D_I2(j+1/*j-1*/,i);
	m_i      = D_M2(j+1/*j-1*/,i);
      }else{
	d_i      = INFINITY;
	m_i      = INFINITY;
      }

      /* compute  d_i(j+1) */
      D_I2(j+1,i) = MIN(
			MIN( ((i) ? D_M2(j,i-1) : INFINITY ), D_M2(j,i)) + gp_cost_gap_opened
			,
			MIN( ((i) ? D_I2(j,i-1) : INFINITY ), D_I2(j,i)) + gp_cost_gap_continued
		       );

      /*  compute des m_i(j+1) */
      D_M2(j+1,i) = MIN ( 
			 m_i_prev  + D_S2(-i,(-j+i-1)), 
			 d_i_prev  + D_S2(-i,(-j+i-1))
			 );
      
      
      xdrop_current = MIN(xdrop_current,D_M2(j+1,i)); 
      
    }/* i */
    if (j&1)
	xdrop_current2 = MIN(xdrop_current,xdrop_prev);
    
    /* printf("XDROP %i\n",xdrop_current2); */
    if (xdrop_current2 >= XDROP){
      /* set the parameters beeing returned */
      return 1;
    }else{
      if(xdrop_current < xdrop_best){
	xdrop_best = xdrop_current;
	*(p_gainscore) = xdrop_best;
	for(i=0;i<=j+1;i++)
	  if (D_M2(j+1,i) == xdrop_best){
	    *(p_last_pos1) = pos1 - i        ;
	    *(p_last_pos2) = pos2 - j + i - 1;
	    break;
	  }
      }
    }  
  }
  return 0;
}




#ifdef INLINE
 inline 
#endif
 int right_alignment_SG(char * data1, int pos1, int len1,
			char * data2, int pos2, int len2, 
			int XDROP, 
			/*out*/ int * p_gainscore,/*out*/ int * p_last_pos1,/*out*/ int * p_last_pos2)
{
  int i,j;
  int xdrop_current,xdrop_prev,xdrop_current2,xdrop_best=INFINITY;
  int d_i_prev,d_i;
  int m_i_prev,m_i;
  
  /* (1) initialisation */
  for(i=0;i<MIN(len1,len2);i++){
    D_I2(0,i) = INFINITY;
    D_I2(1,i) = INFINITY;
    D_M2(0,i) = INFINITY;
    D_M2(1,i) = INFINITY;
  }
 
  D_M2(0,0) = 0;   
  *(p_gainscore) = 0;
  *(p_last_pos1) = pos1;
  *(p_last_pos2) = pos2;
  
  xdrop_current2 = 0;
  xdrop_current  = INFINITY;
  xdrop_best     = 0;
  
  /* (2) compute (stop if Xdrop) */
  for(j=0;j<MIN(len1,len2);j++){/* for each diagonal */
    
    d_i = INFINITY;
    m_i = INFINITY;
    
    xdrop_prev    = xdrop_current;
    xdrop_current = INFINITY;
    
    for(i=0;i<=j+1;i++){
       
      d_i_prev = d_i;
      m_i_prev = m_i;
      
      if (i<j){
	d_i      = D_I2(j+1/*j-1*/,i);
	m_i      = D_M2(j+1/*j-1*/,i);
      }else{
	d_i      = INFINITY;
	m_i      = INFINITY;
      }

      /* printf("j = %i, i = %i :  \t m_i = %i:%i \t d_i = %i:%i \n",j,i,m_i,m_i_prev,d_i,d_i_prev); */
      
      
      /* compute d_i(j+1) */
      D_I2(j+1,i) = MIN(
			MIN( ((i) ? D_M2(j,i-1) : INFINITY ), D_M2(j,i)) + gp_cost_gap_opened
			,
			MIN( ((i) ? D_I2(j,i-1) : INFINITY ), D_I2(j,i)) + gp_cost_gap_continued
		       );

      /* compute m_i(j+1) */
      D_M2(j+1,i) = MIN ( 
			 m_i_prev  + D_S2(i,(j-i+1)), 
			 d_i_prev  + D_S2(i,(j-i+1))
			 );
      
      
      xdrop_current = MIN(xdrop_current,D_M2(j+1,i));
    }/* i */
    if (j&1)
	xdrop_current2 = MIN(xdrop_current,xdrop_prev);
    if (xdrop_current2 >= XDROP){
      /* mise en place parametres retour */
      return xdrop_current2;
    }else{
      if(xdrop_current < xdrop_best){
	xdrop_best = xdrop_current;
	*(p_gainscore) = xdrop_best;
	for(i=0;i<=j+1;i++)
	  if (D_M2(j+1,i) == xdrop_best){
	    *(p_last_pos1) = pos1 + i         ;
	    *(p_last_pos2) = pos2 + j - i + 1 ;
	    break;
	  }
      }
    }  
  }
  return 0;
}


#ifdef INLINE
 inline 
#endif	
 int left_alignment_SG_Lz(char * data1, int pos1, int len1,
			  char * data2, int pos2, int len2, 
			  int XDROP, 
			  /*out*/ int * p_gainscore,
			  /*out*/ int * p_last_pos1,
			  /*out*/ int * p_last_pos2){
   int len = MIN(len1,len2) - gp_delta_stat;
   int last_indels = 0, indels = 0;
   int last_i      = 0, i      = 0;
   int gainscore = 0;
   
   while(1){
     if (!indels)
       gainscore = *p_gainscore +  gp_cost_gap_opened + ABS(indels - last_indels) * gp_cost_gap_continued;
     
     for(i = last_i; i<len; i++ ) {
       gainscore += D_S2(-i,-i+indels);
       if (gainscore > XDROP)
	 break;
       if (gainscore < *p_gainscore) {
	 *p_gainscore  = gainscore;
	 *p_last_pos1  = pos1-i;
	 *p_last_pos2  = pos2-i+indels;
	 last_i        = i;
	 last_indels   = indels;
       }
     }/* for i */
    
     if (indels > 0){
       indels = -indels;
     } else {
       indels = -indels+1;
     }
     if (indels > gp_delta_stat)
       break;
   }/* while */
   return 1;
 }





#ifdef INLINE
 inline 
#endif
 int right_alignment_SG_Lz(char * data1, int pos1, int len1,
			   char * data2, int pos2, int len2, 
			   int XDROP, 
			   /*out*/ int * p_gainscore,
			   /*out*/ int * p_last_pos1,
			   /*out*/ int * p_last_pos2)
{
  int len = MIN(len1,len2) - gp_delta_stat;
  int last_indels = 0, indels = 0;
  int last_i      = 0, i      = 0;
  int gainscore = 0;
  
  while(1){
    if (!indels)
      gainscore = *p_gainscore +  gp_cost_gap_opened + ABS(indels - last_indels) * gp_cost_gap_continued;

    for(i = last_i; i<len; i++ ) {
      gainscore += D_S2(i,i+indels);
      if (gainscore > XDROP)
	break;
      if (gainscore < *p_gainscore) {
	*p_gainscore  = gainscore;
	*p_last_pos1  = pos1+i;
	*p_last_pos2  = pos2+i+indels;
	last_i        = i;
	last_indels   = indels;
      }
    }/* for i */
    
    if (indels > 0){
      indels = -indels;
    } else {
      indels = -indels+1;
    }
    if (indels > gp_delta_stat)
      break;
  }/* while */
  return 1;
}




int right_alignment_SG_Border(char * data1, int pos1, int len1, 
			      char * data2, int pos2, int len2,
			      int XDROP, 
			      /*out*/ int * p_gainscore,
			      /*out*/ int * p_last_pos1,
			      /*out*/ int * p_last_pos2)
{
  /* band alignment length anw width*/
  int bandwidth  = gp_delta_stat;
  
 
  /* various switch elements to access information */
  int row1       = 2*bandwidth;
  int row2       = 2*row1;
  
  /* dp tables */
  int * bm0 = (int *)buffer01;
  int * bm1 = (int *)buffer01+row1;
  int * bi0 = (int *)buffer01+row2;
  int * bi1 = (int *)buffer01+row2+row1;
  
  /* maximal and minimal positions taken into account */
  int j,k;
  int jmin = 0, jmax = row1;
  
  for(j=0;j<row1;j++)
    bm0[j] = bi0[j] = (gp_cost_gap_opened) + ABS(j-bandwidth)*gp_cost_gap_continued;
  
  bm0[bandwidth] = bi0[bandwidth] = 0;

  if (len1 < len2){
    
    for(k=1 ; k< len1; k++){ 
     
      if(k&1){	
	/* M(j) full vectorisable loop */
	for(j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data1 + pos1)[k])] [(int)(data2 + pos2)[j+k-bandwidth]];
	  bm1[j] = MIN( bm0[j] , bi0[j] ) + d; 
	  
	  if (bm1[j] < *p_gainscore) {
	    *p_gainscore = bm1[j];
	    *p_last_pos1 = pos1 + k;
	    *p_last_pos2 = pos2 + j + k - bandwidth;
	  }
	  
	}	
	
	/* MinxMax evaluations */
	j=jmin;	
	while( bm1[j++] > XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm1[j--] > XDROP && j>0)
	  jmax--;
	if (jmax<jmin)
	  return XDROP+1;
  

	/* I(j) in two half loops */
	bi1[bandwidth] = MIN(MIN(bi0[bandwidth-1],bi0[bandwidth+1]) + gp_cost_gap_continued , MIN(bm0[bandwidth-1],bm0[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi1[j-1] = MIN(
			 MIN(bi0[j-1],bi1[j]) + gp_cost_gap_continued,
			 MIN(bm0[j-1],bm1[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi1[j] = MIN(
		       MIN(bi0[j],bi1[j-1]) + gp_cost_gap_continued,
		       MIN(bm0[j],bm1[j-1]) + gp_cost_gap_opened
		       );
	}

      }else{/* k&1*/
          
      	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data1 + pos1)[k])][(int)(data2 + pos2)[j+k-bandwidth]];
	  bm0[j] = MIN( bm1[j] , bi1[j] ) + d;

	  if (bm0[j] < *p_gainscore) {
	    *p_gainscore = bm0[j];
	    *p_last_pos1 = pos1 + k;
	    *p_last_pos2 = pos2 + j + k - bandwidth;
	  }

	}

	/* MinxMax evaluations */
	j=jmin;	
	while( bm0[j++] > XDROP && j<jmax)
	  jmin++;
	j=jmax;	
	while( bm0[j--] > XDROP && j>0)
	  jmax--;
	if (jmax<jmin)
	  return XDROP+1;
	

	/* I(j) in two half loops */
	bi0[bandwidth] = MIN(MIN(bi1[bandwidth-1],bi1[bandwidth+1]) + gp_cost_gap_continued , MIN(bm1[bandwidth-1],bm1[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi0[j-1] = MIN(
			 MIN(bi1[j-1],bi0[j]) + gp_cost_gap_continued,
			 MIN(bm1[j-1],bm0[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi0[j] = MIN(
		       MIN(bi1[j],bi0[j-1]) + gp_cost_gap_continued,
		       MIN(bm1[j],bm0[j-1]) + gp_cost_gap_opened
		       ); 
	}
      }/* k&1*/
    }/* k */

  }else{
    
    for( k=1 ; k< len2; k++){ 
          
      if(k&1){	
	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[((int)(data1 + pos1)[j+k-bandwidth])][(int)(data2 + pos2)[k]];
	  bm1[j] = MIN( bm0[j] , bi0[j] ) + d;

	  if (bm1[j] < *p_gainscore) {
	    *p_gainscore = bm1[j];
	    *p_last_pos1 = pos1 + k;
	    *p_last_pos2 = pos2 + j + k - bandwidth;
	  }

	}	

	/* MinxMax evaluations */
	 j=jmin;	
	while( bm1[j++] > XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm1[j--] > XDROP && j>0)
	  jmax--;
	
	if (jmax<jmin)
	  return XDROP+1;

	/* I(j) in two half loops (each on is vectorisable on "non stupid" compilers) */
	bi1[bandwidth] = MIN(MIN(bi0[bandwidth-1],bi0[bandwidth+1]) + gp_cost_gap_continued , MIN(bm0[bandwidth-1],bm0[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi1[j-1] = MIN(
			 MIN(bi0[j-1],bi1[j]) + gp_cost_gap_continued,
			 MIN(bm0[j-1],bm1[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi1[j] = MIN(
		       MIN(bi0[j],bi1[j-1]) + gp_cost_gap_continued,
		       MIN(bm0[j],bm1[j-1]) + gp_cost_gap_opened
		       );
	}

      }else{/* k&1*/
          
      	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data2 + pos2)[k])<<2][(int)(data1 + pos1)[j+k-bandwidth]];
	  bm0[j] =  MIN( bm1[j] , bi1[j] ) + d; 

	  if (bm0[j] < *p_gainscore) {
	    *p_gainscore = bm0[j];
	    *p_last_pos1 = pos1 + k;
	    *p_last_pos2 = pos2 + j + k - bandwidth;
	  }

	}	

	/* MinxMax evaluations */
	j=jmin;
	while( bm0[j++] > XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm0[j--] > XDROP && j>0)
	  jmax--;
	
	if (jmax<jmin)
	  return XDROP+1;
	
	/* I(j) in two half loops (each on is vectorisable on "non stupid" compilers) */
	bi0[bandwidth] = MIN(MIN(bi1[bandwidth-1],bi1[bandwidth+1]) + gp_cost_gap_continued , MIN(bm1[bandwidth-1],bm1[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi0[j-1] = MIN(
			 MIN(bi1[j-1],bi0[j]) + gp_cost_gap_continued,
			 MIN(bm1[j-1],bm0[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi0[j] = MIN(
		       MIN(bi1[j],bi0[j-1]) + gp_cost_gap_continued,
		       MIN(bm1[j],bm0[j-1]) + gp_cost_gap_opened
		       ); 
	}
      }/* k&1*/
    }/* k */
 
  }
  return (MAX(len1,len2)&1)?(bm0[bandwidth]):(bm1[bandwidth]);
}





int left_alignment_SG_Border(char * data1, int pos1, int len1, 
			     char * data2, int pos2, int len2,
			     int XDROP, 
			      /*out*/ int * p_gainscore,
			      /*out*/ int * p_last_pos1,
			      /*out*/ int * p_last_pos2)
{
  /* band alignment length anw width*/
  int bandwidth  = gp_delta_stat;
  
 
  /* various switch elements to access information */
  int row1       = 2*bandwidth;
  int row2       = 2*row1;
  
  /* dp tables */
  int * bm0 = (int *)buffer01;
  int * bm1 = (int *)buffer01+row1;
  int * bi0 = (int *)buffer01+row2;
  int * bi1 = (int *)buffer01+row2+row1;
  
  /* maximal and minimal positions taken into account */
  int j,k;
  int jmin = 0, jmax = row1;
  
  for(j=0;j<row1;j++)
    bm0[j] = bi0[j] = (gp_cost_gap_opened) + ABS(j-bandwidth)*gp_cost_gap_continued;
  
  bm0[bandwidth] = bi0[bandwidth] = 0;

  if (len1 < len2){
    
    for(k=1 ; k< len1; k++){ 
     
      if(k&1){	
	/* M(j) full vectorisable loop */
	for(j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data1 + pos1)[-k])][(int)(data2 + pos2)[j-k-bandwidth]];
	  bm1[j] = MIN( bm0[j] , bi0[j] ) + d; 
	  
	  if (bm1[j] < *p_gainscore) {
	    *p_gainscore = bm1[j];
	    *p_last_pos1 = pos1 - k;
	    *p_last_pos2 = pos2 + j - k - bandwidth;
	  }
	  
	}	
	
	/* MinxMax evaluations */
	j=jmin;	
	while( bm1[j++] > XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm1[j--] > XDROP && j>0)
	  jmax--;
	if (jmax<jmin)
	  return XDROP+1;
  

	/* I(j) in two half loops */
	bi1[bandwidth] = MIN(MIN(bi0[bandwidth-1],bi0[bandwidth+1]) + gp_cost_gap_continued , MIN(bm0[bandwidth-1],bm0[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi1[j-1] = MIN(
			 MIN(bi0[j-1],bi1[j]) + gp_cost_gap_continued,
			 MIN(bm0[j-1],bm1[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi1[j] = MIN(
		       MIN(bi0[j],bi1[j-1]) + gp_cost_gap_continued,
		       MIN(bm0[j],bm1[j-1]) + gp_cost_gap_opened
		       );
	}

      }else{/* k&1*/
          
      	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data1 + pos1)[-k])][(int)(data2 + pos2)[j-k-bandwidth]];
	  bm0[j] = MIN( bm1[j] , bi1[j] ) + d;

	  if (bm0[j] < *p_gainscore) {
	    *p_gainscore = bm0[j];
	    *p_last_pos1 = pos1 - k;
	    *p_last_pos2 = pos2 + j - k - bandwidth;
	  }

	}

	/* MinxMax evaluations */
	j=jmin;	
	while( bm0[j++] > XDROP && j<jmax)
	  jmin++;
	j=jmax;	
	while( bm0[j--] > XDROP && j>0)
	  jmax--;
	if (jmax<jmin)
	  return XDROP+1;
	

	/* I(j) in two half loops */
	bi0[bandwidth] = MIN(MIN(bi1[bandwidth-1],bi1[bandwidth+1]) + gp_cost_gap_continued , MIN(bm1[bandwidth-1],bm1[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi0[j-1] = MIN(
			 MIN(bi1[j-1],bi0[j]) + gp_cost_gap_continued,
			 MIN(bm1[j-1],bm0[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi0[j] = MIN(
		       MIN(bi1[j],bi0[j-1]) + gp_cost_gap_continued,
		       MIN(bm1[j],bm0[j-1]) + gp_cost_gap_opened
		       ); 
	}
      }/* k&1*/
    }/* k */

  }else{
    
    for( k=1 ; k< len2; k++){ 
          
      if(k&1){	
	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data2 + pos2)[-k])][(int)(data1 + pos1)[j-k-bandwidth]];
	  bm1[j] = MIN( bm0[j] , bi0[j] ) + d;

	  if (bm1[j] < *p_gainscore) {
	    *p_gainscore = bm1[j];
	    *p_last_pos1 = pos1 - k;
	    *p_last_pos2 = pos2 + j - k - bandwidth;
	  }

	}	

	/* MinxMax evaluations */
	 j=jmin;	
	while( bm1[j++] > XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm1[j--] > XDROP && j>0)
	  jmax--;
	
	if (jmax<jmin)
	  return XDROP+1;

	/* I(j) in two half loops (each on is vectorisable on "non stupid" compilers) */
	bi1[bandwidth] = MIN(MIN(bi0[bandwidth-1],bi0[bandwidth+1]) + gp_cost_gap_continued , MIN(bm0[bandwidth-1],bm0[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi1[j-1] = MIN(
			 MIN(bi0[j-1],bi1[j]) + gp_cost_gap_continued,
			 MIN(bm0[j-1],bm1[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi1[j] = MIN(
		       MIN(bi0[j],bi1[j-1]) + gp_cost_gap_continued,
		       MIN(bm0[j],bm1[j-1]) + gp_cost_gap_opened
		       );
	}

      }else{/* k&1*/
          
      	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data2 + pos2)[-k])][(int)(data1 + pos1)[j-k-bandwidth]];
	  bm0[j] =  MIN( bm1[j] , bi1[j] ) + d; 

	  if (bm0[j] < *p_gainscore) {
	    *p_gainscore = bm0[j];
	    *p_last_pos1 = pos1 - k;
	    *p_last_pos2 = pos2 + j - k - bandwidth;
	  }

	}	

	/* MinxMax evaluations */
	j=jmin;
	while( bm0[j++] > XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm0[j--] > XDROP && j>0)
	  jmax--;
	
	if (jmax<jmin)
	  return XDROP+1;
	
	/* I(j) in two half loops (each on is vectorisable on "non stupid" compilers) */
	bi0[bandwidth] = MIN(MIN(bi1[bandwidth-1],bi1[bandwidth+1]) + gp_cost_gap_continued , MIN(bm1[bandwidth-1],bm1[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi0[j-1] = MIN(
			 MIN(bi1[j-1],bi0[j]) + gp_cost_gap_continued,
			 MIN(bm1[j-1],bm0[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi0[j] = MIN(
		       MIN(bi1[j],bi0[j-1]) + gp_cost_gap_continued,
		       MIN(bm1[j],bm0[j-1]) + gp_cost_gap_opened
		       ); 
	}
      }/* k&1*/
    }/* k */
 
  }
  return (MAX(len1,len2)&1)?(bm0[bandwidth]):(bm1[bandwidth]);
}





/*
 *  [II]
 *
 *   "slow" alignment fucntions used to compute statistical
 *   alignements parameters.
 *
 ***********************************************************************/




#define M3(i,j) (m[(i)+ (j)*(len1+1)])
#define I3(i,j) (in[(i)+ (j)*(len1+1)])
#define D3(i,j) (d[(i)+ (j)*(len1+1)])
#define S3(i,j) (S2(i,j))

#define M_TO_M_SUBS  129
#define I_TO_M_SUBS  131
#define FIRST_DEL_LEFT  133
#define FIRST_DEL_RIGHT 135
#define CONT_DEL_LEFT  137
#define CONT_DEL_RIGHT 139

int alignment_SG_stats(char * data1, int pos1, int len1,
		       char * data2, int pos2, int len2,
		       /*out*/int    * trans       /*[2]*/,
		       /*out*/int    * stats_data1 /*[3]*/,
		       /*out*/int    * stats_data2 /*[3]*/){
  
  int i,j;
  int i1,i2;
  int result;

  int * m = (int*) malloc( (unsigned) (len1+1) * (len2+1) * sizeof(int));
  int * in = (int*) malloc( (unsigned) (len1+1) * (len2+1) * sizeof(int));  
  int * d = (int*) malloc( (unsigned) (len1+1) * (len2+1) * sizeof(int));
  
  ASSERT(m,alignment_SG_stats);
  ASSERT(in,alignment_SG_stats);
  ASSERT(d,alignment_SG_stats);
  
  /* [1] compute */

  M3(0,0) = 0;
  
  I3(0,0) = gp_cost_gap_opened - gp_cost_gap_continued;
  
  for(i=1;i<len1+1; i++)
    M3(i,0) = INFINITY;
  
  for(i=1 ; i<len1+1 ; i++)
    I3(i,0) = I3(i-1,0) + gp_cost_gap_continued;
  

  for(j=1 ; j<len2+1 ; j++){
  
    M3(0,j) =  INFINITY;
    I3(0,j) =  gp_cost_gap_opened  + (j-1)*gp_cost_gap_continued;
    
    for(i=1;i<len1+1;i++){
      I3(i,j) = MIN (
		     MIN( M3(i,j-1) + gp_cost_gap_opened    , M3(i-1,j) + gp_cost_gap_opened )
		     ,
		     MIN( I3(i,j-1) + gp_cost_gap_continued , I3(i-1,j) + gp_cost_gap_continued )
		     );
      M3(i,j) = MIN ( M3(i-1,j-1) + S3(i,j), I3(i-1,j-1) + S3(i,j) ); 
    }
  }
  
  result = MIN( M3(len1,len2) , I3(len1,len2) );
 
  /* [2] backtrace */
 
  i = len1;
  j = len2;

  if (M3(len1,len2) > I3(len1,len2))
    D3(i,j) = I_TO_M_SUBS;
  else
    D3(i,j) = M_TO_M_SUBS;

  
  while (i > 0 && j > 0){
        
    if ( D3(i,j) == M_TO_M_SUBS || D3(i,j) == FIRST_DEL_LEFT || D3(i,j) == FIRST_DEL_RIGHT ){

      /* no gap to open  */

      if ( M3(i,j) == M3(i-1,j-1) + S3(i,j) ){ 
	i--;
	j--;
	D3(i,j) = M_TO_M_SUBS;
	}else{	
	  if ( M3(i,j) == I3(i-1,j-1) + S3(i,j) ){ 
	    i--;
	    j--;
	    D3(i,j) = I_TO_M_SUBS;
	  }else{
	    WARNING("alignment_SG_stats() 1");
	  }
	}
    }else{
      
      /* need to open a gap */
      if ( I3(i,j) == M3(i,j-1) + gp_cost_gap_opened ){
	j--;
	D3(i,j) =  FIRST_DEL_RIGHT ;
      }else{
	if ( I3(i,j) == M3(i-1,j) + gp_cost_gap_opened ){
	  i--;
	  D3(i,j) =  FIRST_DEL_LEFT ;
	}else{
	  if ( I3(i,j) == I3(i,j-1) + gp_cost_gap_continued ){
	    j--;
	    D3(i,j) =  CONT_DEL_RIGHT ;
	  }else{
	    if ( I3(i,j) == I3(i-1,j) + gp_cost_gap_continued ){
	      i--;
	      D3(i,j) =  CONT_DEL_LEFT ;
	    }else{
	      WARNING("alignment_SG_stats() 2");
	    }
	  } 
	}
      }
    }
  }/* while */
  
    
  while(j > 0){
    j--;
    D3(0,j) = CONT_DEL_RIGHT;
    D3(0,0) = FIRST_DEL_RIGHT;
  }
  
  while(i > 0){
    i--;
    D3(i,0) = CONT_DEL_LEFT;
    D3(0,0) = FIRST_DEL_LEFT;
  }
  
  free(m);
  free(in);


  /* [3] backtrace to compute some statistics */

  i1 = 0; i2 = 0; 
 
 
  while (i1 < len1 || i2 < len2 ){
    
    switch(D3(i1,i2)){
    case M_TO_M_SUBS:
    case I_TO_M_SUBS:
      
      if (data1[pos1+i1]!=data2[pos2+i2]){ /* substitution */
	
	(stats_data1[(pos1+i1)%3])++;
	(stats_data2[(pos2+i2)%3])++;
	
	if (((data1[pos1+i1])^(data2[pos2+i2])) == 2)
	  (trans[0])++; /* transition */
	else
	  (trans[1])++; /* transversion */
      }
      i1++;
      i2++;
      
    break;
	
    case FIRST_DEL_LEFT:
    case CONT_DEL_LEFT:
      trans[2]++;
      i1++;
      break;
      
    case FIRST_DEL_RIGHT:
    case CONT_DEL_RIGHT:
      trans[2]++;
      i2++;
      break;
      
    default:
       WARNING("alignment_SG_stats() 3");
      
    } 
    
  }/* while */

  free(d);
  return result; 
}


int alignment_SG_stats_Strait(char * data1,  int pos1, char * data2,  int pos2, int len,
			      /*out*/int    * trans       /*[2]*/,
			      /*out*/int    * stats_data1 /*[3]*/,
			      /*out*/int    * stats_data2 /*[3]*/){
  int i;
  for(i=0;i<len;i++){ 
    if (data1[pos1+i]!=data2[pos2+i]){ /* substitution */
	(stats_data1[(pos1+i)%3])++;
	(stats_data2[(pos2+i)%3])++;
	if (((data1[pos1+i])^(data2[pos2+i])) == 2)
	  (trans[0])++; /* transition */
	else
	  (trans[1])++; 
    }
  }
  return 0;
}




int alignment_SG_stats_on_MA(char * data1, char * data2,
			     MA * ma,
			     /*out*/int * trans       /*[2]*/,
			     /*out*/int * stats_data1 /*[3]*/,
			     /*out*/int * stats_data2 /*[3]*/){
  int left_size,right_size;
  tuple * t = ma->first_tuple, * t_prev = ma->first_tuple ;
  
  /* [1] Compute alignment before first tuple */
  
  left_size  = TBL_POS(t) - ma->left_pos_begin  + 1;
  right_size = TBR_POS(t) - ma->right_pos_begin + 1;
  
  if ( left_size > 0 && right_size > 0)
    alignment_SG_stats(data1, ma->left_pos_begin , left_size ,
		       data2, ma->right_pos_begin, right_size,
		       trans,stats_data1,stats_data2);
   
  /* [2] Compute alignment for several tuples */
 
  while(t != NULL){
    /* (1) compute inside a tuple */
    alignment_SG_stats_Strait(data1,TBL_POS(t),
			      data2,TBR_POS(t),TSIZE(t),
			      trans,stats_data1,stats_data2);
    if (t->next){
      /* (2) alignment between (t) et (t->next) */
      alignment_SG_stats(data1,TEL_POS(t),TGAP_L(t,t->next),
			 data2,TER_POS(t),TGAP_R(t,t->next),
			 trans,stats_data1,stats_data2);
    }/* if */
    t_prev = t;
    t = t->next;
  }/* while */

  /* [3] Compute alignement after last tuple */
  
  left_size  =  ma->left_pos_end  - TEL_POS(t_prev) + 1;
  right_size =  ma->right_pos_end - TER_POS(t_prev) + 1;
  
  
  if ( left_size > 0 && right_size> 0)
    alignment_SG_stats(data1, TEL_POS(t_prev), left_size ,
		       data2, TER_POS(t_prev), right_size,
		       trans,stats_data1,stats_data2);
  
  return 0;
}



/*
 *  [III]
 *
 *   "slow"  alignment methods used to visualise
 *   provided alignments 
 *
 ***********************************************************************/



/*
 *
 *  display buffer composed of 5 line with NBL chars
 *
 */


int line_pos = 0;
char line_dpos1[NBL+1];
char line_d1[NBL+1];
char line_op[NBL+1];
char line_d2[NBL+1];
char line_dpos2[NBL+1];



void flush()
{
  int i;
  if (line_pos > 0){ 
    /* close strings */
    i = line_pos;
    if (line_dpos1[i]!='\0' ){
      while (line_dpos1[i]!=' ' && i>0)
	i--;
      line_dpos1[i]='\0';
    }
    
    i = line_pos;
    if (line_dpos2[i]!='\0'){
      while (line_dpos2[i]!=' ' && i>0)
	i--;
      line_dpos2[i]='\0';
    }
			       
    line_dpos1[line_pos] = '\0';
    line_dpos2[line_pos] = '\0';
    line_d1[line_pos]    = '\0';
    line_d2[line_pos]    = '\0';
    line_op[line_pos]    = '\0';
    /* print the 5 lines */
    fprintf(OUTSTREAM,line_dpos1);fprintf(OUTSTREAM,"\n");
    fprintf(OUTSTREAM,line_d1   );fprintf(OUTSTREAM,"\n");
    fprintf(OUTSTREAM,line_op   );fprintf(OUTSTREAM,"\n");
    fprintf(OUTSTREAM,line_d2   );fprintf(OUTSTREAM,"\n");
    fprintf(OUTSTREAM,line_dpos2);fprintf(OUTSTREAM,"\n");
    fprintf(OUTSTREAM,"\n");
    /* reset buffers */
    line_pos = 0;
    memset(line_dpos1,0,NBL+1);
    memset(line_dpos2,0,NBL+1);
    memset(line_d1   ,0,NBL+1);
    memset(line_d2   ,0,NBL+1);
    memset(line_op   ,0,NBL+1);
  }
}

void push(char d1,int dpos1,char op,char d2, int dpos2)
{
  /* put positions in the buffer */
  if ( !(dpos1%10)  && size_int(dpos1) + line_pos < NBL){
    if (line_pos>0)
      if (line_dpos1[line_pos-1] == '|')
	line_dpos1[line_pos-1] = ' ';
      sprintf(line_dpos1+line_pos,  "|%i",dpos1);
  }
  if (line_dpos1[line_pos]=='\0')
    line_dpos1[line_pos]=' ';
  
  
  if (!(dpos2%10)   && size_int(dpos2) + line_pos < NBL){
    if (line_pos>0)
      if (line_dpos2[line_pos-1] == '|')
	line_dpos2[line_pos-1] = ' ';
    sprintf(line_dpos2+line_pos,"|%i",dpos2);
  }
  if (line_dpos2[line_pos]=='\0')
    line_dpos2[line_pos]=' ';
 
  
  /* memorize alignment */
  
  line_d1[line_pos]=d1;
  line_d2[line_pos]=d2;  
  line_op[line_pos]=op;
  
  line_pos++;

  /* print if end of line */
  if (line_pos == NBL){
    flush();
  } 
}


/*************************************************************/





/* SIMPLE ALIGNEMENT 
 * $ GAP1
 * $ SUB1
 */


#define M(i,j) (m[(i)+ (j)*(len1+1)])
#define D(i,j) (d[(i)+ (j)*(len1+1)])

#define DIAG_EGAL  121
#define DIAG_DIFF  123
#define STEP_LEFT  125
#define STEP_RIGHT 127

 int display_alignment_noflush(int reverse,
				     char * data1, int pos1, int len1,
				     char * data2, int pos2, int len2)
{
  
  int i,j,diag;
  int i1,i2;
  
  int * m = (int*) malloc( (unsigned) (len1+1) * (len2+1) * sizeof(int));
  int * d = (int*) malloc( (unsigned) (len1+1) * (len2+1) * sizeof(int));
  
  ASSERT(m,display_alignment_noflush);
  ASSERT(d,display_alignment_noflush);

  /* [1] compute */
  M(0,0) = 0;
  
  for(j=0; j < len2; j++)
    M(0,j+1) = M(0,j) + GAP1; 
  
  for(i=0; i < len1; i++)                           
    { 
      M(i+1,0) = M(i,0)  + GAP1; 
      for(j=0; j < len2 ; j++)                        
	{ 
	  diag = M(i,j);
	  if( data1[pos1+i] != data2[pos2+j]) 
	    diag+=SUB1;
	  M(i+1,j+1) = MIN3 ( diag, M(i,j+1) + GAP1 , M(i+1,j) + GAP1);
	  
	}		
    }	
  
  
  /* [2] backtrace */ 
  i = len1;
  j = len2;
  while (i > 0 && j > 0){
      if ( M(i,j) ==  M(i-1,j)  + GAP1 ){
	i--;
	D(i,j) = STEP_LEFT;
      }else{ 
	if ( M(i,j) ==  M(i,j-1) + GAP1){
	  j--;
	  D(i,j) = STEP_RIGHT;
	}else{
	  if ( M(i,j) == M(i-1,j-1) && (data1[pos1+i-1] == data2[pos2+j-1])){ 
	    i--;
	    j--;
	    D(i,j) = DIAG_EGAL;
	  }else{ 
	    if ( M(i,j) ==  M(i-1,j-1) + SUB1 && (data1[pos1+i-1] != data2[pos2+j-1])){
	      j--;i--;
	      D(i,j) = DIAG_DIFF;
	    }else{
	      WARNING("display_alignment_noflush() 1");
	    }
	  }
	}
      }
  }
  
  while(j > 0)
    D(0,--j) = STEP_RIGHT;
  while(i > 0)
    D(--i,0) = STEP_LEFT;
 
  free(m);
  
  

  /* [3] display alignment */
  i1 = 0; i2 = 0; 
 
  while (i1 < len1 || i2 < len2 ) {
    
    switch(D(i1,i2)){
	
      case DIAG_EGAL:
	push(LOOKUP(data1[pos1+i1]),
	     BCOUNT + ((reverse) ? (gp_querysize_chunk - 1) - (pos1 + i1) : pos1+i1),
	     CHAR_EQU,
	     LOOKUP(data2[pos2+i2]),
	     BCOUNT + pos2 + i2
	     );
	i1++;i2++;
	break;
	
      case DIAG_DIFF:
	push(LOOKUP(data1[pos1+i1]),
	     BCOUNT + ((reverse) ? (gp_querysize_chunk - 1)- (pos1 - i1) : pos1+i1),
	     (((data1[pos1+i1])^(data2[pos2+i2]))==2)?CHAR_SS:CHAR_SV,
	     LOOKUP(data2[pos2+i2]),
	     BCOUNT + pos2 + i2
	     );
	
	i1++;i2++;
	break;
	
      case STEP_LEFT:
	push(LOOKUP(data1[pos1+i1]),
	     BCOUNT + ((reverse) ? (gp_querysize_chunk - 1) - (pos1 + i1) : pos1+i1),
	     CHAR_DEL,
	     CHAR_DUM,
	     BCOUNT + pos2 + i2
	     );
	i1++;
	break;
	
      case STEP_RIGHT:
	push(CHAR_DUM,
	     BCOUNT + ((reverse) ? (gp_querysize_chunk - 1) - (pos1 + i1) : pos1+i1),
	     CHAR_INS,
	     LOOKUP(data2[pos2+i2]),
	     BCOUNT + pos2+i2
	     );
	i2++;
	break;
	
      default:
	WARNING("display_alignment_noflush() 2");
      }
     
  }/* while */

  free(d);
  return 0;
}



 int display_alignment(int reverse,
			     char * data1, int pos1, int len1,
			     char * data2, int pos2, int len2)
{
  display_alignment_noflush(reverse,
			    data1,pos1,len1,
			    data2,pos2,len2);
  flush();
  return 0;  
}


 int display_alignment_on_MA(char * data1, char * data2,
				   MA * ma)
     
{
  int i;
  int left_size,right_size;
  tuple * t = ma->first_tuple, * t_prev = ma->first_tuple;

  /* [1] Display alignment before first tuple */
  left_size  = TBL_POS(t) - ma->left_pos_begin  + 1;
  right_size = TBR_POS(t) - ma->right_pos_begin + 1;
  
  if ( left_size > 0 && right_size > 0)
    display_alignment_noflush(ma->reverse,
			      data1, ma->left_pos_begin , left_size ,
			      data2, ma->right_pos_begin, right_size
			      );

  /* [2] Display alignment for several tuples */
  while(t != NULL){

    /* (1) display alignment of tuple "t" */
    for(i = 0 ;i < gp_seed_span + (t->size) ; i++){
      push(LOOKUP(data1[i+TBL_POS(t)]),
	   BCOUNT + ((ma->reverse) ? (gp_querysize_chunk - 1) - (i + TBL_POS(t)) : i+TBL_POS(t)),
	   (data1[i+TBL_POS(t)] == data2[i+TBR_POS(t)])?CHAR_EQU:((data1[i+TBL_POS(t)]^data2[i+TBR_POS(t)])==2)?CHAR_SS:CHAR_SV,
	   LOOKUP(data2[i+TBR_POS(t)]),
	   BCOUNT + i+TBR_POS(t)
	   );      
    }

    
    if (t->next){
      /* (2) display alignment between (t) and (t->next) */
      display_alignment_noflush(ma->reverse,
				data1,TEL_POS(t),TGAP_L(t,t->next),
				data2,TER_POS(t),TGAP_R(t,t->next));
    }
    t = t->next;
    t_prev = t;
  }

  /* [3] Display alignment after first tuple */
  
  left_size  =  ma->left_pos_end  - TEL_POS(t_prev) ;
  right_size =  ma->right_pos_end - TER_POS(t_prev) ;

  if ( left_size > 0 && right_size> 0)
    display_alignment_noflush(ma->reverse,
			      data1, TEL_POS(t_prev), left_size ,
			      data2, TER_POS(t_prev), right_size
			      );
  flush();
 
  return 0;
}


/* ALIGNEMENT WITh SUBS MATRIX + GAP OPEN CONTINUE
 * gp_cost_gap_opened
 * gp_cost_gap_continued
 * substitution_matrix
 */



#define M3(i,j) (m[(i)+ (j)*(len1+1)])
#define I3(i,j) (in[(i)+ (j)*(len1+1)])
#define D3(i,j) (d[(i)+ (j)*(len1+1)])

#define M_TO_M_SUBS  129
#define I_TO_M_SUBS  131
#define FIRST_DEL_LEFT  133
#define FIRST_DEL_RIGHT 135
#define CONT_DEL_LEFT  137
#define CONT_DEL_RIGHT 139


 int display_alignment_SG_noflush(int reverse,
					char * data1, int pos1, int len1,
					char * data2, int pos2, int len2)
{
  
  int i,j;
  int i1,i2;
  
  
  int * m = (int*) malloc( (unsigned) (len1+1) * (len2+1) * sizeof(int));
  int * in = (int*) malloc( (unsigned) (len1+1) * (len2+1) * sizeof(int));  
  int * d = (int*) malloc( (unsigned) (len1+1) * (len2+1) * sizeof(int));
  
  ASSERT(m,display_alignment_SG_noflush);
  ASSERT(in,display_alignment_SG_noflush);
  ASSERT(d,display_alignment_SG_noflush);
  

  /* [1] compute */
  
  M3(0,0) = 0; 
  I3(0,0) = gp_cost_gap_opened - gp_cost_gap_continued;
  
  for(i=1 ; i<len1+1 ; i++)
    M3(i,0) = INFINITY;
 
  for(i=1 ; i<len1+1 ; i++)
    I3(i,0) = I3(i-1,0) +  gp_cost_gap_continued;
  
  for(j=1 ; j<len2+1 ; j++){
    
    M3(0,j) =  INFINITY;
    I3(0,j) =  I3(0,j-1)  + gp_cost_gap_continued;
    
    for(i=1;i<len1+1;i++){
      I3(i,j) = MIN (
		     MIN( M3(i,j-1) + gp_cost_gap_opened    , M3(i-1,j) + gp_cost_gap_opened )
		     ,
		     MIN( I3(i,j-1) + gp_cost_gap_continued , I3(i-1,j) + gp_cost_gap_continued )
		     );
      M3(i,j) = MIN ( M3(i-1,j-1) + S3(i,j), I3(i-1,j-1) + S3(i,j) ); 
    }
  }
 
  
  /* [2] backtrace */
 
  i = len1;
  j = len2;

  if (M3(len1,len2) > I3(len1,len2))
    D3(i,j) = I_TO_M_SUBS;
  else
    D3(i,j) = M_TO_M_SUBS;

  
  while (i > 0 && j > 0){
        
    if ( D3(i,j) == M_TO_M_SUBS || D3(i,j) == FIRST_DEL_LEFT || D3(i,j) == FIRST_DEL_RIGHT ){

      /* no gap to open */

      if ( M3(i,j) == M3(i-1,j-1) + S3(i,j) ){ 
	i--;
	j--;
	D3(i,j) = M_TO_M_SUBS;
	}else{	
	  if ( M3(i,j) == I3(i-1,j-1) + S3(i,j) ){ 
	    i--;
	    j--;
	    D3(i,j) = I_TO_M_SUBS;
	  }else{
	    WARNING("display_alignment_SG_noflush() 1");
	  }
	}
    }else{
      
      /* need to open a gap */
      if ( I3(i,j) == M3(i,j-1) + gp_cost_gap_opened ){
	j--;
	D3(i,j) =  FIRST_DEL_RIGHT ;
      }else{
	if ( I3(i,j) == M3(i-1,j) + gp_cost_gap_opened ){
	  i--;
	  D3(i,j) =  FIRST_DEL_LEFT ;
	}else{
	  if ( I3(i,j) == I3(i,j-1) + gp_cost_gap_continued ){
	    j--;
	    D3(i,j) =  CONT_DEL_RIGHT ;
	  }else{
	    if ( I3(i,j) == I3(i-1,j) + gp_cost_gap_continued ){
	      i--;
	      D3(i,j) =  CONT_DEL_LEFT ;
	    }else{
	      WARNING("display_alignment_SG_noflush() 2");
	    }
	  } 
	}
      }
    }
  }/* while */
  
    
  while(j > 0){
    j--;
    D3(0,j) = CONT_DEL_RIGHT;
    D3(0,0) = FIRST_DEL_RIGHT;
  }
  
  while(i > 0){
    i--;
    D3(i,0) = CONT_DEL_LEFT;
    D3(0,0) = FIRST_DEL_LEFT;
  }
  
  free(m);
  free(in);
  /* [3] display alignment */

  i1 = 0; i2 = 0; 
 
 
  while (i1 < len1 || i2 < len2 ){
    
    switch(D3(i1,i2)){
    case M_TO_M_SUBS:
    case I_TO_M_SUBS:
      push(LOOKUP(data1[pos1+i1]),
	   BCOUNT + ((reverse) ? (gp_querysize_chunk - 1) - (pos1 + i1) : pos1+i1),
	   (data1[pos1+i1]==data2[pos2+i2])?CHAR_EQU:((((data1[pos1+i1])^(data2[pos2+i2]))==2)?CHAR_SS:CHAR_SV),
	   LOOKUP(data2[pos2+i2]),
	   BCOUNT + pos2+i2
	   );
      i1++;
      i2++;
    break;
	
    case FIRST_DEL_LEFT:
    case CONT_DEL_LEFT:
      push(LOOKUP(data1[pos1+i1]),
	   BCOUNT + ((reverse) ? (gp_querysize_chunk - 1) - (pos1 + i1) : pos1+i1),
	   CHAR_DEL,
	   CHAR_DUM,
	   BCOUNT + pos2+i2
	   );
      i1++;
      break;
      
    case FIRST_DEL_RIGHT:
    case CONT_DEL_RIGHT:
      push(CHAR_DUM,
	   BCOUNT + ((reverse) ? (gp_querysize_chunk - 1) - (pos1 + i1) : pos1+i1),
	   CHAR_INS,
	   LOOKUP(data2[pos2+i2]),
	   BCOUNT + pos2+i2
	   );
      i2++;
      break;
      
    default:
      WARNING("display_alignment_SG_noflush() 3");
    } 
    
  }/* while */
  free(d);

  return 0;  
}





 int display_alignment_SG(int reverse,
			  char * data1, int pos1, int len1,
			  char * data2, int pos2, int len2)
{
  display_alignment_SG_noflush(reverse,
			       data1,pos1,len1,
			       data2,pos2,len2);
  flush();
  return 0;  
}


 int display_alignment_SG_on_MA(char * data1,char * data2,
				MA * ma)
			    
{
  int i;
  int left_size,right_size;
  tuple * t = ma->first_tuple, * t_prev = ma->first_tuple;
  
  /* [1] Display alignment before first tuple */
  
  left_size  = TBL_POS(t) - ma->left_pos_begin  ;
  right_size = TBR_POS(t) - ma->right_pos_begin ;
  
  if ( left_size > 0 && right_size > 0)
    display_alignment_SG_noflush(ma->reverse,
			      data1, ma->left_pos_begin , left_size ,
			      data2, ma->right_pos_begin, right_size
			      );
  
  /* push('#',1,'#','#',1); */

  /* [2] Display alignment for several tuples */
  
  while(t != NULL){

    /* (1) display alignment of tuple "t" */     
    for(i = 0 ;i < gp_seed_span + (t->size) ; i++){
      push(LOOKUP(data1[i+TBL_POS(t)]),
	   BCOUNT + ((ma->reverse) ? (gp_querysize_chunk - 1) - (i + TBL_POS(t)) : i+TBL_POS(t)),
	   (data1[i+TBL_POS(t)] == data2[i+TBR_POS(t)])?CHAR_EQU:((data1[i+TBL_POS(t)]^data2[i+TBR_POS(t)])==2)?CHAR_SS:CHAR_SV,
	   LOOKUP(data2[i+TBR_POS(t)]),
	   BCOUNT + i + TBR_POS(t)
	   );      
    }

    
    if (t->next){
      /* (2) display alignment between (t) and (t->next) */
      display_alignment_SG_noflush(ma->reverse,
				   data1,TEL_POS(t),TGAP_L(t,t->next),
				   data2,TER_POS(t),TGAP_R(t,t->next));
    }
    t_prev = t;
    t = t->next;
  }

  /* [3] Display alignment after first tuple */
 
  left_size  =  ma->left_pos_end  - TEL_POS(t_prev) ;
  right_size =  ma->right_pos_end - TER_POS(t_prev) ;
 
  /* push('#',1,'#','#',1); */
  
  if ( left_size > 0 && right_size> 0)
    display_alignment_SG_noflush(ma->reverse,
				 data1, TEL_POS(t_prev), left_size ,
				 data2, TER_POS(t_prev), right_size);
 

  flush();
  return 0;
}

