/*
 *  YASS 1.07
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "util.h"

/*****************************
 *
 * Waiting Time Distribution
 *
 *****************************/


/*
 * bound of waiting time
 *  
 */


#ifdef INLINE
 inline 
#endif
int statistical_bound_of_waiting_time1(double p,int k, double alpha)
{
  double a1  =  dpow(p,k) ;
  double a2  = (1-p) *  a1;
  int    x   = 0;
  double Sum = 0.00;

  double * last_k_prob = (double *)malloc((unsigned)(k+1)*sizeof(double));
  ASSERT(last_k_prob,waiting_time_distrib_1);
    
  
  while( Sum < (1-alpha)){
    
    if (x < k){
      last_k_prob[x%(k+1)] = 0.00;
    }else if (x == k){
      last_k_prob[x%(k+1)] = a1;
    }else if (x <= 2*k){
      last_k_prob[x%(k+1)] = a2;
    }else{
      last_k_prob[x%(k+1)] = last_k_prob[(x+k)%(k+1)]- a2*last_k_prob[x%(k+1)];
    }
    Sum += last_k_prob[x%(k+1)];
    x++;
  }
  free(last_k_prob);
  
  return x;
}


#ifdef INLINE
 inline 
#endif
int statistical_bound_of_waiting_time2(double p,int k, double alpha)
{
  double p_k   =  dpow(p,k) ;
  double qp_k  = (1-p) *  p_k;
  int    x   = 0;
  double Sum_0_xk1 = 0.00;
  double Sum = 0;
  
  double * last_k_prob = (double *)malloc((unsigned)(k+1)*sizeof(double));
  ASSERT(last_k_prob,waiting_time_distrib_2);
  
  while( Sum < (1-alpha)){
    if (x < k){
      last_k_prob[x%(k+1)] = 0.00;
    }else if (x == k){
      last_k_prob[x%(k+1)] = p_k;
    }else {
      Sum_0_xk1 += last_k_prob[x%(k+1)];
      last_k_prob[x%(k+1)] = qp_k *(1 - Sum_0_xk1);
    }
    
    Sum += last_k_prob[x%(k+1)];
    x++;
  }
  free(last_k_prob);
  
  return x;
}




/*****************************
 *
 * Random Walk distribution:
 *
 *****************************/


/*
 *
 * table of probabilities to reach 
 * position "i" at step L  
 * with the rdwalk  
 * -> pr pI for left shift
 * -> pr pI for right shift
 * (pr 1-2pI dont move) 
 */


#ifdef INLINE
 inline 
#endif
double * randomwalk_probability_of_pos1(double pI,int L){

  double * C_N_K      ;/* Cnk table */
  double * P_for_score;/* random walk scores */ 
  double Sum_of_C_N_K = 1;
  double P_dep;
  
  int i,nb_dep;
  
  /* (1) allocation and initialisation */
  
  C_N_K = (double *)malloc((unsigned)(2*L+3)*sizeof(double));  
  ASSERT(C_N_K,randomwalk_probability_of_pos1);
  for(i=0;i<2*L+3;i++)
    C_N_K[i] = 0;
  
  
  P_for_score = (double *)malloc((unsigned)(2*L+1)*sizeof(double));  
  ASSERT(P_for_score,randomwalk_probability_of_pos1);
  for(i=0;i<2*L+1;i++)
    P_for_score[i] = 0;

  

  /* (2) poly computed
   */
  
  C_N_K[L] = 1;


  for(nb_dep=0 ; nb_dep <=L ; nb_dep ++){
        
    P_dep = C(L,nb_dep) * dpow(pI,nb_dep) * dpow((1-pI),(L-nb_dep));
     
    
    /***************************************/
    if (P_dep > 1 || 1){ /* infinity problems ??? */
      goto end;
    }
   

    /* (2.1) proba computed here */
    for(i=-nb_dep;i<=nb_dep;i++){
      P_for_score[i+L] += ((C_N_K[i+L]) * P_dep / Sum_of_C_N_K);
    }

    /* (2.2) C_N_K update */
    if (nb_dep<L){
      for(i=-nb_dep-1;i<=nb_dep+1;i+=2){
	C_N_K[i+L] = C_N_K[i-1+L] + C_N_K[i+1+L];
      }
      for(i=-nb_dep;i<=nb_dep;i+=2){
	C_N_K[i+L] = 0;
      }
      Sum_of_C_N_K *= 2; 
    }
  }
  
 end:
  /* (3) free unused stuff*/
  free (C_N_K);
  
  return P_for_score;
}


#ifdef INLINE
 inline 
#endif
double * randomwalk_probability_of_pos2(double pI,int L){
  
  int i,j;
  double a = pI / 2;
  double b = 1 - pI;
  double * poly    = (double *)malloc((unsigned)(2*L+1)*sizeof(double));  
  double * polytmp = (double *)malloc((unsigned)(2*L+1)*sizeof(double));
  
  ASSERT(poly,randomwalk_probability_of_pos2);
  ASSERT(polytmp,randomwalk_probability_of_pos2);


  /* (1)  table inits */
  
  for(i=0;i<2*L+1;i++){
    poly[i]=0;
  }
  
  /* (2) polynom computed
   * [ p*X^2 + b*X + p ] ^ L
   *
   */
  
  poly[0] = 1;
  for(i=0;i<L;i++){
    
    /* clear tmp buffer */
    for(j=0;j<(2*L+1);j++){
      polytmp[j]=0;
    }

    /* badly have [ p*X^2 + b*X + p ] ^ i+1 from [ p*X^2 + b*X + p ] ^ i */
    for(j=0;j<(2*L+1);j++)
      polytmp[j]+=poly[j]*a;
    for(j=0;j<(2*L);j++)
      polytmp[j+1]+=poly[j]*b;
    for(j=0;j<(2*L-1);j++)
      polytmp[j+2]+=poly[j]*a;
    
    /* copy [ p*X^2 + b*X + p ] ^ i+1 into poly[] */
    for(j=0;j<(2*L+1);j++)
      poly[j]=polytmp[j];
    
  }/* i */
  
  free(polytmp);
  
  return poly;
  
}


#ifdef INLINE
 inline 
#endif
double * randomwalk_probability_of_pos3(double pI,int L){
   
   int i,j;
   double a = pI * 0.50;
   double b = 1 - pI;
   double * u = (double *)malloc((unsigned)(2*L+1)*sizeof(double));
   double * f = (double *)malloc((unsigned)(2*L+1)*sizeof(double));
   double * t = (double *)malloc((unsigned)(2*L+1)*sizeof(double));
   double * s = NULL;
   int P = L;
   
   
  ASSERT(u , randomwalk_probability_of_pos3);
  ASSERT(f , randomwalk_probability_of_pos3);
  ASSERT(t , randomwalk_probability_of_pos3);
  
  /* (1) tables inits */ 
 
  for(i = 0 ; i < 2*L+1 ; i++)
    u[i] = 0;
  for(i = 0 ; i < 2*L+1 ; i++)
    f[i] = 0;
  
  f[0] = 1.00;
  u[0] = a;
  u[1] = b;
  u[2] = a;

  /* (2) qpow */
  while(P > 0) {
    if (P&1){
      /* f = f*u */ 
      for(i=0 ; i<2*L+1 ; i++){
	t[i] = 0;
	for(j=0 ; j<=i ; j++)
	  t[i] += u[i-j] * f[j];
      } 
      s = t;
      t = f;
      f = s;
    }
    /* u = u * u; */
    for(i=0 ; i<2*L+1 ; i++){
      t[i] = 0;
      for(j=0 ; j<=i ; j++)
	t[i] += u[i-j] * u[j];
    }
    s = t;
    t = u;
    u = s;
    P>>=1;    
  }
  
  free(u);
  free(t);
  return f;
  
}







/*
 *
 * give rd walk bounds in  (1-alpha)% 
 * 
 */



#ifdef INLINE
 inline 
#endif
int statistical_bound_of_randomwalk1(double pI,int L,double alpha)
{
  double * RDW_Bound = randomwalk_probability_of_pos1(pI,L);
  double Sum         = RDW_Bound[L];
  int    bound       = 0;
  while(Sum < (1-alpha) && bound < L){    
    bound++;
    Sum += RDW_Bound[L-bound] + RDW_Bound[L + bound];
  }

  /* overflow in table*/
  if (bound ==  L)
    bound = (int) (pow((double)pI*L,(double)(1/2.3))) + 1; /* asympt value */
  
  free(RDW_Bound);  
  return bound;
}




#ifdef INLINE
 inline 
#endif
int statistical_bound_of_randomwalk2(double pI,int L,double alpha)
{
  double * RDW_Bound = randomwalk_probability_of_pos3(pI,L);
  double Sum         = RDW_Bound[L];
  int    bound       = 0;
  while(Sum < (1-alpha) && bound < L){    
    bound++;
    Sum += RDW_Bound[L-bound] + RDW_Bound[L + bound];
  }
  
  free(RDW_Bound);
  return bound;
}



/*
 * Scores,Evalue,and Bitscore ...
 *
 */

double Evalue(double K, double Lambda, int query_size, int text_size, int score)
{
  return K * query_size * text_size * exp(- Lambda * score);
}

double BitScore(double K, double Lambda, int score)
{
  return (Lambda * ((double)score) - log(K))/log(2);
}

int MinScore(double K, double Lambda,  int query_size, int text_size, double MaxEvalue)
{
  return (int) floor((1/Lambda)*(log(K*query_size*text_size) - log(MaxEvalue)));
}



/*
 *  trinomial prob
 */

#ifdef INLINE
 inline 
#endif
double P_mutation_bias(int freq[])
{
  int i;
  int n = freq[0] + freq[1] + freq[2];
  double result = 1.00;

  if (n == 0)
    return 1.00;

  for(i=1;i<=n;i++){
    result *= ((double)i)/3;
  }
  for(i=1;i<=freq[0];i++)
    result /= (double)i;
  for(i=1;i<=freq[1];i++)
    result /= (double)i;  
  for(i=1;i<=freq[2];i++)
    result /= (double)i; 
  
  return result;
}

