/*
 *  YASS 1.07
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "tuple.h"
#include "global_var.h"
#include "util.h"
#include "regroup.h"

/*** GLOBAL DECLARATIONS ***/
int gv_windows = 0;



/*We use a tree which can be a red black tree or an avl tree*/

#ifdef CHOOSERBTREE
   rb_tree *T;
   #define TREE_TYPE(x) rb##x
#else
#ifdef CHOOSEAVLTREE
   avl_tree *T;
   #define TREE_TYPE(x) avl##x
#else
#error No tree definition : uncomment either "CHOOSERBTREE" or "CHOOSEAVLTREE" inside  "util.h"
#endif
#endif


/*The global queue of MA, ordered by right pos end*/
queue_MA *Q;

/*The current position (start of the window)*/
int   win_position;
float left = 1;
float right = 1;


/*
* int size_queue_MA (queue_MA *q)
*
* Calculate the number of elements in a queue
*/

int size_queue (queue_MA *q) {

   int i = 0;
   list_MA *l = q->first;
   while (l) {
      l = l -> next;
      i++;
   }
   return i;
}


/*
* int deepness_max (TREE_TYPE(_node) *n)
*
* Calculate the maximal depth of a tree
*/

int deepness_max(TREE_TYPE(_node) *n) {

   int h1;
   int h2;
   if (n == NULL) {
     return 0;
   } else {
     h1 = deepness_max(n->TREE_TYPE(_link[0]));
     h2 = deepness_max(n->TREE_TYPE(_link[1]));     
     return (1 + MAX(h1,h2));
   }
   
}
  
  
/*
* nb_max_elem_queue (TREE_TYPE(_node *n))
*
* Calculate the maximum of MA in a node list 
*/  
  
int nb_max_elem_queue(TREE_TYPE(_node *n)) {

   int nb = 0;
   int nb1;
   int nb2;
   
   list_MA *l;
   
   if (n == NULL) {return 0;}
   
   else {
   
      nb1 = nb_max_elem_queue(n -> TREE_TYPE(_link[0]));
      nb2 = nb_max_elem_queue(n -> TREE_TYPE(_link[1]));
      
      l = (n->data).queue->first;
      
      while (l) {
      
         l = l -> next;
         nb++;
	 
      }
      
      return (MAX3(nb,nb1,nb2));
      
   }
   
}


/*
* elem nb_aver_elem_queue (TREE_TYPE(_node *n))
*
* Calculate the average number of MA in all the node lists of a tree 
*/  

elem nb_aver_elem_queue (TREE_TYPE(_node *n)) {

   elem e;
   elem e1;
   elem e2;
   
   list_MA *l;
   
   if (n == NULL) {
   
      e.size = 0;
      e.weight = 0;
      return e;
      
   }
   else {
   
      e1 = nb_aver_elem_queue(n -> TREE_TYPE(_link[0]));
      e2 = nb_aver_elem_queue(n -> TREE_TYPE(_link[1]));
      
      l = (n->data).queue->first;
      
      e.size = 0;
      
      while (l) {
      
         l = l -> next;
         e.size++;
	 
      }
   
      e.weight = 1 + e1.weight + e2.weight;
      e.size = (e.size + e1.weight * e1.size + e2.weight * e2.size) / (e.weight);
      
      return e;
      
   }
      
}

int ratio_left_right(TREE_TYPE(_node) *n,int h) {

   float i = 1/h;
   int n1;
   int n2;
   
   if (n == NULL) {return 0;}
   
   else {
   
      n1 = ratio_left_right (n->TREE_TYPE(_link)[0],h+1);
      n2 = ratio_left_right (n->TREE_TYPE(_link)[1],h+1);
      
      left = left + i * n1;
      right = right + i * n2;
      
      return (1 + MAX(n1,n2));
      
   }
   
}
   









/*
* void TREE_TYPE(_stats ())
*
* Print several statistical results
*/


void TREE_TYPE(_stats) () {

   ratio_left_right (T->root,1);
   
   printf("%d\t%d\t%d\t%d\t%d\t%d\t%f\t%f\n",win_position,
	                                 nb_postprocessed_grouping_tests,
					 nb_postprocessed_grouping_links,
					 size_queue(Q),
	                                 deepness_max(T->root),
					 nb_max_elem_queue(T->root),
					 (nb_aver_elem_queue(T->root)).size,
					 right/left);
					
}
   

/*
* TREE_TYPE(_node * smallest (TREE_TYPE(_tree) *T)
*
* Return the smallest element af a tree
*/

TREE_TYPE(_node) * smallest (TREE_TYPE(_tree) *T) {

   TREE_TYPE(_node) *y = T->root;
   TREE_TYPE(_node) *s = NULL;
   
   while (y != NULL) {
   
      s = y;
      y = y -> TREE_TYPE(_link[0]);
	 
   }
  
   return s;
   
}


/*
* prod_and_customer * rb_q_delete (rb_tree *tree, MA *ma)
*
* Delete a node from the tree
*/

list_MA * tree_q_delete (TREE_TYPE(_tree) *tree, MA *ma) {

   tree_data         *data          = NULL;
   list_MA           *list_MA_curr  = NULL;

   assert (tree != NULL && ma != NULL);

   /*we search the element in the tree*/
   data = TREE_TYPE(_search)(tree,DIAGB(ma));

   if (data){
     /* when found we delete it from its queue in the tree */
     list_MA_curr = queue_extract_MA(data->queue,ma);
      
     /* after that, if the queue is empty we delete the node */
     if (data->queue->first == NULL) {      
       free_data(data);
       TREE_TYPE(_delete_rebalancing)(tree); 
     }
   }else{
     WARNING("rb_q_delete() searched element not found : data empty");
   }
   return list_MA_curr;
}





/*
* void display_tree_list (TREE_TYPE(_tree *T))
*
* Display the ordered list of nodes of a tree 
*/

void display_tree_list (TREE_TYPE(_tree) *T) {

   TREE_TYPE(_node) *y = smallest(T);
   tree_data *data = NULL;
   
   list_MA *l;
   int i = 0;
   
   if (y) {
     data = &(y->data);
   } else {
     printf("+++ Empty List ! +++\n");
   }
   
   while (data) {
     l = data->queue->first; 
     i = 0;
     while(l) {i++;l = l -> next;}
     printf("(dist=%d, elems=%d)",data->distance,i);
     data = data->list[1];
   }
   printf("\n");
   
}

void verif_tree_table (TREE_TYPE(_node) *n) {

   list_MA *l; 
   
   if (n == NULL)
     return;
   
   l = (n -> data) . queue -> first;
   
   printf("*******************************\n");
   
   while(l) {
   
      printf("%x %x %x\n",(int)l,(int)l->producer,(int)l->customer);
   
      if (l->customer == NULL) {
      
         printf("ERREUR CUSTOMER !!!\n");
	 
      }
      
      if (l->producer == NULL) {
      
         printf("ERREUR PRODUCER !!!\n");
	 
      }

      l = l -> next;
      
   }
   
   verif_tree_table(n->TREE_TYPE(_link)[0]);
   verif_tree_table(n->TREE_TYPE(_link)[1]);

}


void clear_tree () {

   list_MA *l;
   list_MA *pac;

   while (T -> root) {
     l = (T->root->data) . queue->first;

     pac = queue_extract_MA(Q,l->ma);
     if (pac) 
       free(pac); 
     else 
       WARNING("clear_tree() : missing prod_and_cust in Q ");

     pac = tree_q_delete(T,l->ma);
     if (pac){
       if (pac->customer) free(pac->customer);
       if (pac->producer) free(pac->producer);
       free(pac);
     } else 
       WARNING("clear_tree() : missing prod_and_cust in T ");
   }
   free(T);
   free(Q);
}


/*
* void window_deleter (int pos);
*
* Delete all the MA which are no longer in the windows at position "pos"  
*/

#ifdef INLINE 
inline
#endif
void window_deleter (int pos) {

   list_MA *itself;
   MA  *current_MA;
   MA  *target_MA;
   
   tuple *t;
   tuple *tr;
   
   list_MA *list_MA_left; 
   list_MA *list_MA_right; 
   list_MA *pac;
   
   list_MA *first_customer = NULL; 
   list_MA *first_producer = NULL;
   
   /* data and data_sn are temporary variable */
   tree_data *data;
   tree_data *data_sn;
   
   list_MA *liste;
  
   int canbegrouped = 0;
   
   /* while elements are in the window */
   
   while ( (Q->first != NULL) && (Q->first->ma->right_pos_end < pos)) {
      
     current_MA = Q->first->ma;
      
     /* we delete the element from the tree and the global queue */
     list_MA_left = tree_q_delete(T,current_MA);						      
     free(queue_pop_MA(Q));
      
     /* we search for a customer of the deleted MA which best producer is the deleted MA */ 
     if (list_MA_left){
      
       int i;
       canbegrouped = 0; 
       if( list_MA_left->customer->index > 0){
	 first_customer = list_MA_left   -> customer -> table[0];
	 first_producer = first_customer -> producer -> table[0];
	 if ( first_producer == list_MA_left ){
	   canbegrouped = 1;
	 }
       }
       
       /* if MAs can be regrouped together */	
       if (canbegrouped) { 
	 target_MA = first_customer->ma;
	  
#ifdef STATS
	 nb_postprocessed_grouping_links++;
#endif	 
	  
	 /*we delete the regrouped MA from the tree*/
	 list_MA_right = tree_q_delete(T,target_MA);
	 
	 /* FIXME */ assert (list_MA_right->ma==target_MA);
	  
	  
  
	 if (list_MA_right){
	   
	   /*we delete the regrouped MA from the global queue*/
	   pac = queue_extract_MA(Q,target_MA);
	   if (pac) 
	     free(pac); 
	   else 
	     WARNING("window_deleter() :  missing \"prod_and_cust\" in target_MA");
	   
	   
	   /* we signal to other customer from the left MA that it won't be no more a producer */
	   
	   for (i=1 ; i<list_MA_left->customer->index ; i++) {
	     list_MA * i_customer = list_MA_left->customer->table[i];
	     table_MA_delete(i_customer->producer, list_MA_left);
	   }
	  
	   /* we signal to other producer from the right MA that it won't be no more a customer */
	   for (i=1 ; i<list_MA_right->producer->index; i++) {
	     list_MA * i_producer = list_MA_right->producer->table[i];
	     table_MA_delete(i_producer->customer, list_MA_right);
	   }
	   
	   /* We build the new MA */
	   
	   
	   /*FIXME : give exact computation */
	   if ( target_MA->blastscore + DIAG_COST(DIAG_LENGTH(current_MA,target_MA)) + INDEL_COST(INDEL_LENGTH(DIAGE(current_MA),DIAGB(target_MA))) > 0)
	     current_MA->blastscore   += target_MA->blastscore + DIAG_COST(DIAG_LENGTH(current_MA,target_MA)) + INDEL_COST(INDEL_LENGTH(DIAGE(current_MA),DIAGB(target_MA))); 
	   else /* Indel approx */
	     if ( target_MA->blastscore + INDEL_COST(INDEL_LENGTH(DIAGE(current_MA),DIAGB(target_MA))) > 0)
	       current_MA->blastscore   += target_MA->blastscore +  INDEL_COST(INDEL_LENGTH(DIAGE(current_MA),DIAGB(target_MA))); 
	   
	   current_MA->left_pos_end  = target_MA->left_pos_end;
	   current_MA->right_pos_end = target_MA->right_pos_end;
	    
	   t = current_MA->first_tuple;
	   tr = target_MA->first_tuple;
	  
	   while ((t->next) && (TBR_POS(t->next)<TBR_POS(tr)) && (TBL_POS(t->next)<TBL_POS(tr)))
	     t = t->next;
	    
	   while ((TER_POS(t) >= TBR_POS(tr)) || (TEL_POS(t) >= TBL_POS(tr))) {
	     t->occurrence--;
	     t->size--;
	    }
	  
	   t->next             = tr;
	   target_MA->first_tuple = NULL;
	  
#ifdef DEBUG 
	   memset(list_MA_left->customer,0,sizeof(table_MA));
	   memset(list_MA_right->producer,0,sizeof(table_MA));
#endif
	  
	   free(list_MA_left->customer);
	   free(list_MA_right->producer);
	  
	   /* then we push it in order for it to be the last of the list */
	   
	   itself = queue_push_and_sort_MA(Q,current_MA,NULL);
	   
	   /* and then we insert it into the tree */
	   data_sn = TREE_TYPE(_find)(T,DIAGB(current_MA));
	   data    = TREE_TYPE(_insert)(T,DIAGB(current_MA));
	   
	   if (data->queue != NULL) {
	     liste = queue_push_and_sort_MA(data->queue,current_MA,itself);
	   } else {
	     liste = new_data(data,data_sn,itself);
	   }
	  
	   /* we give back to liste the producer of the left MA and the customer of the right MA */
	   liste -> customer = list_MA_right->customer;
	   liste -> producer = list_MA_left->producer;
	   
	   /* we signal to the customer of liste that its adress has changed */
	   
	   for ( i=0 ; i < liste->customer->index ; i++) {
	     list_MA * i_customer = liste->customer->table[i];
	     table_MA_delete(i_customer->producer, list_MA_right);
	     assert (table_MA_insert(i_customer->producer,liste,current_MA->blastscore,i_customer) == 1);
	   }
	   
	   /* we try to find new producers for list */
	   regroup(liste);
	   
#ifdef DEBUG 
	   memset(list_MA_left,0,sizeof(list_MA));
	   memset(list_MA_right,0,sizeof(list_MA));
	   
#endif
	   free(list_MA_left);
	   free(list_MA_right);
	 }else{
	   WARNING("window_deleter () : missing \"list_MA_right\" in T ");
	   goto leftcannotbegrouped;
	 }
       } else {
       leftcannotbegrouped:
	 /* If the MA can't be regrouped we signal to its customer taht it can't be no more a producer */
	 for (i=0;i< list_MA_left->customer->index; i++) {
	   list_MA * i_customer = list_MA_left->customer->table[i];
	   table_MA_delete(i_customer->producer, list_MA_left);
	 }
	 
#ifdef DEBUG 
	 memset(list_MA_left->customer,0,sizeof(table_MA));
	 memset(list_MA_left->producer,0,sizeof(table_MA));
	 memset(list_MA_left,0,sizeof(list_MA));	  
#endif	
	 
	 free(list_MA_left->producer);
	 free(list_MA_left->customer);
	 free(list_MA_left);
	 
       }
     }else{
       WARNING("window_deleter () : missing \"list_MA_left\" in T ");
     }
   }
}


/*
 * int regroup (MA *item);
 *
 * Test with blascore and distance criterions if a MA can be linked with an 
 * other MA and proceed if possible
 */

#ifdef INLINE 
inline
#endif    
void regroup (list_MA *liste) {
  
  /*y is the current node*/
  tree_data *y = TREE_TYPE(_find) (T,DIAGB(liste->ma)); 
  /*y_plus is y->right*/
  tree_data *y_plus = NULL;
  /*best is the nearest node from liste->ma between y and y_plus*/ 
  tree_data *best = NULL;
  
  /*d is the value of the distance to diagonale of the begin of liste->ma*/
  int d = DIAGE(liste->ma);
  
  /*min and max are the bound between which we can make a regroupement*/
  int min = d - INDEL_BOUND(liste->ma);
  int max = d + INDEL_BOUND(liste->ma);
  
  /*if hit == 1 then we make a regroupement*/
  int hit = 0;
  
  /*the list_MA which is regrouped with liste*/
  list_MA *l = NULL;
  
  /*First we choose which data will be the most efficient to start our comparisons*/
  
  y_plus = y -> list[1];
  best = y;
  
  /*At this point we start the search of an MA which can be regroup with liste->ma*/
  
  /*While potential candidates do exist*/
  
  while (((y != NULL) || (y_plus != NULL)) && (!hit)) {
    
    l = best -> queue -> last;
    
    /*For each element in the node queue which are valuable for linking*/
    
    while ((l) && (l->ma->blastscore + l->ma->right_pos_end > liste->ma->right_pos_begin) && (!hit)) {
      
      /*if hit == 1 then we have found the good MA*/
      
      hit = GOOD_LINK(l->ma->blastscore, 
		      liste->ma->blastscore,
		      BLAST(l->ma->blastscore,
			    liste->ma->blastscore,
			    INDEL_COST(INDEL_LENGTH(best->distance,d)),
			    DIAG_COST(DIAG_LENGTH(l->ma,liste->ma))));
      
      
      
      /*********************************************************************************************/
      /*  ATTENTION CECI EST PROVISOIRE ET N'EST UTILE QUE POUR EVITER LES BUGS DE L'OPTION -d 1  */
      /*********************************************************************************************/
      
      if ((liste->ma->right_pos_begin - l->ma->right_pos_end) > 5000) {hit = 0;}
      
      /*********************************************************************************************/
      
      if (l->ma == liste->ma) {hit = 0;}
      
      
      /* if the two MA can be regrouped*/
      
      if (hit) {
	
	/*we insert the new MA in the customer table of the old MA*/
	
	if (table_MA_insert(l->customer,liste,hit,l)) {
	  
	  /*if this insertion has succeeded we insert the old MA in the producer table of the new MA*/
	  
	  if (!table_MA_insert(liste->producer,l,hit,liste)) {
	    
	    /*if this insertin has failed we delete the new MA from the customer table of the old MA*/
	    
	    table_MA_delete(l->customer,liste);
	    
	  }     
	  
	}
	
	hit = 0;
	
      }
      
#ifdef STATS         	       
      nb_postprocessed_grouping_tests++;
#endif
      
      l = l -> prev;
      
    }
    
    if (!hit) {
      
      /*we change y or y_plus depending on who was the best*/
      
      if (best == y) {
	
	y = y -> list[0];
	
      }
      else {
	
	y_plus = y_plus -> list [1];
	
      }
      
      /*we test if we are under the bounds*/
      
      if ((y != NULL) && (y -> distance < min)) {
	
	y = NULL;
	
      }
      
      if ((y_plus != NULL) && (y_plus -> distance > max)) {
	
	y_plus = NULL;
	
      }
      
      /*then we decide who is the best between y and y_plus (if we go on the left or on the right of the list)*/
      
      if (y != NULL) {
	
	best = y;
	
	if ((y_plus != NULL) && (y_plus->distance - d) < (d - y->distance)) {
	  
	  best = y_plus;
	  
	}
	
      }
      else {
	
	if (y_plus != NULL) {
	  
	  best = y_plus;
	  
	}
	
      }
      
    }
    
  }
  
}


/*
 * MA * main_regroup (MA *ma, int r);
 *
 * Proceed the global regroupement (r means "reverse") 
 */

#ifdef INLINE 
inline
#endif
MA * main_regroup (MA *ma, int r) {
  
  /*the current MA*/
  MA * my_MA = ma;
  
  /*the prevedent MA*/
  MA * prev_MA;
  MA * to_free_MA;
  
  int i_chunk = my_MA->i_chunk;
  
  tree_data *data;
  tree_data *data_sn; /*sn = smallest nearest*/
  
  list_MA *itself;
  list_MA *l;
  
  int old_nb_link = -1;
  
  /*while we can make new regroupement*/
  
  while(old_nb_link < nb_postprocessed_grouping_links) {
    
    my_MA = ma;
    
    old_nb_link =  nb_postprocessed_grouping_links;
    
    T = malloc(sizeof(TREE_TYPE(_tree)));
    T->root = NULL;
    
    Q = malloc(sizeof(queue_MA));
    Q->first = NULL;
    Q->last  = NULL;
    
    win_position = 0;
    
    /*for each MA which have the good reverse field and the same i_chunk*/
    
    while ((my_MA) && (my_MA->reverse == r) && (my_MA->i_chunk == i_chunk)) {
      
      /*we insert it into the global queue*/
      
      itself = queue_push_and_sort_MA(Q,my_MA,NULL);
      
      /*we insert it into the tree*/
      
      data_sn = TREE_TYPE(_find)(T,DIAGB(my_MA));
      data    = TREE_TYPE(_insert)(T,DIAGB(my_MA));
      
      /*we fill the tree_data field*/
      
      if (data->queue != NULL) {
	
	l=queue_push_and_sort_MA(data->queue,my_MA,itself);
	l->customer = malloc(sizeof(table_MA));
	l->producer = malloc(sizeof(table_MA));
	table_MA_init(l->customer);
	table_MA_init(l->producer);
	
      } else {
	
	l=new_data(data,data_sn,itself);
	
      }
      
      /*we try to regroup the MA*/
      
      regroup(l); 
      
      /*if the new MA is out of the window we move it*/
      
      if (my_MA -> right_pos_begin > (win_position + gv_windows)) {
	
	win_position = my_MA -> right_pos_begin - gv_windows;
	
	/*then we delete the elements which are now out of the window*/
	
	window_deleter(win_position);
	
      }
      
#ifdef DEBUG
      TREE_TYPE(_stats) ();
#endif
      
      my_MA = my_MA->next;
#ifdef STATS	
      nb_postprocessed_ma++;
#endif	
    }
    
    /*we push the window on the right border*/
    win_position = gp_textsize;
    window_deleter(win_position);
    
    clear_tree ();
    
    /*At this point we delete all the MA which have been regrouped*/  
    
    my_MA = ma;
    prev_MA = NULL;
    
    while ((my_MA) && (my_MA->reverse == r) && (my_MA->i_chunk == i_chunk)) {
      
      if (my_MA -> first_tuple == NULL) {
	
	to_free_MA = my_MA;
	
	if (prev_MA != NULL) {
	  
	  prev_MA -> next = my_MA -> next;
	  
	}
	
	my_MA = my_MA -> next;
	free(to_free_MA);
	
      }
      else {
	
	prev_MA = my_MA;
	my_MA = my_MA -> next;
	
      }
      
    }
    
  }
  
  return my_MA;
  
}         


void main_multi_regroup (MA *ma) {
  
  MA *current_MA = ma;
  MA *the_MA = NULL;
  
  gv_windows = gp_win_min;
  
  if (gp_win_mul > 1) {
    if (gp_reverse == 2) {
      while ((current_MA) && (current_MA->reverse == 0)) {
	while (gv_windows <= gp_win_max) {
	  the_MA = main_regroup(current_MA,0);
	  gv_windows = gv_windows * gp_win_mul;
	}
	current_MA = the_MA;
	gv_windows = gp_win_min;
      }
      
      while (current_MA) {
	while (gv_windows <= gp_win_max) {
	  the_MA = main_regroup(current_MA,1);
	  gv_windows = gv_windows * gp_win_mul;
	}
	current_MA = the_MA;
	gv_windows = gp_win_min;
      }
    }
    else {
      while (current_MA) {
	while (gv_windows <= gp_win_max) {
	  the_MA = main_regroup(current_MA,gp_reverse);
	  gv_windows = gv_windows * gp_win_mul;
	}
	current_MA = the_MA;
	gv_windows = gp_win_min;
      }
    }
  }
}	 





























