/*
 *  YASS 1.07
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include "tuple.h"
#include "display.h"
#include "global_var.h"
#include "prdyn.h"
#include "proba.h"
#include "kword.h"

#ifdef INLINE
 inline 
#endif
tuple     * CreateTuple    (int occurrence, int distance)
{
  
#ifdef GORE_ALLOC_TUPLE
  tuple * t = NULL; 
  while((tupler[i_tuple]).distance != 0){
    i_tuple++;
    i_tuple%=GORE_ALLOC_TUPLE_SIZE;
    if (!i_tuple)
      fprintf(stderr,"#");
  }
  t = &(tupler[i_tuple]);
#else
  tuple * t = (tuple *) malloc((unsigned)sizeof(tuple));
  ASSERT(t,CreateTuple);
#endif

  t->occurrence       = occurrence;
  t->distance         = distance;   
  t->size             = 0 ;
  t->bitweight        = gp_seed_bitweight;  
  t->next             = NULL;
  t->overlappedhits   = 0x00000001;
  return t;
}


#ifdef INLINE
 inline 
#endif
tuplelist * CreateTupleList()
{  

#ifdef GORE_ALLOC_TUPLELIST  
  tuplelist * tl = NULL;
  while((tuplelister[i_tuplelist]).first_tuple != NULL){
    i_tuplelist++;
    i_tuplelist%=GORE_ALLOC_TUPLELIST_SIZE;
    if (!i_tuplelist)
      fprintf(stderr,"*");
  }
  tl = &(tuplelister[i_tuplelist]);
#else
  tuplelist * tl = (tuplelist *) malloc((unsigned)sizeof(tuplelist));
  ASSERT(tl,CreateTupleList);
#endif
  tl->first_tuple = NULL;
  tl->next        = NULL;  
  return tl;
}



#ifdef INLINE
 inline 
#endif
int KeepTupleList(tuplelist * tl,tuple * tfirst, tuple * tlast, MA * ma)
{
#ifdef  GORE_ALLOC_TUPLE
  tuple * t      = tl->first_tuple->next;
  tuple * t_prev = tl->first_tuple;
   
  if (tfirst == tl->first_tuple){  
    tl->first_tuple = tlast->next;
  }else{
    while (t != tfirst){
      t_prev = t ;
      t = t->next;
    }
    t_prev->next = tlast->next;
  }
  ma->first_tuple = tfirst;
  tlast->next     = NULL;
  
#else

  tuple * t      = tl->first_tuple->next;
  tuple * t_prev = tl->first_tuple;  
  if (tfirst == tl->first_tuple){  
    tl->first_tuple = tlast->next;
  }else{
    while (t != tfirst){
      t_prev = t ;
      t = t->next;
    }
    t_prev->next = tlast->next;
  }
  ma->first_tuple = tfirst;
  tlast->next     = NULL;

#endif  
  return 0;
}



#ifdef INLINE
 inline 
#endif
void FreeTuple(tuple * t)
{
#ifdef GORE_ALLOC_TUPLE
  t->distance = 0;
#else
  free(t);
#endif
}



/* only used in "Align_Single" and "Align_Double" */

#ifdef INLINE
 inline 
#endif
void FreeTupleList(tuplelist ** p_tl,tuplelist ** p_tl_prev)
{
  tuple * t, * t_prev;

  if (gv_last_tl == *(p_tl))
    gv_last_tl = *(p_tl_prev);

  t = (*(p_tl))->first_tuple;
  while (t != NULL){
    t_prev = t ;
    t = t->next;
    FreeTuple(t_prev);
  } 
  
  (*(p_tl_prev))->next = (*(p_tl))->next;
#ifdef GORE_ALLOC_TUPLELIST 
  (*(p_tl))->first_tuple = NULL;
#else	
  free(*(p_tl));
#endif
  *(p_tl) = (*(p_tl_prev))->next;}


#ifdef INLINE
 inline 
#endif
MA *  CreateMA(char * data_query  , int datasize_query ,
	       char * data_text   , int datasize_text  ,
	       int left_pos_begin , int right_pos_begin,
	       int left_pos_end   , int right_pos_end  ,
	       tuplelist * tuple_list,
	       tuple * first_tuple_aligned,
	       tuple * last_tuple_aligned,
	       int blastscore,
	       /* 	
		* int mss, int Mss, int nbs,
		*/  
	       int reverse,int i_chunk)
{

  MA * ma = (MA * )malloc((unsigned) sizeof(MA));
  ASSERT(ma,memorize_alignment);

  /* add  "ma" in the MA_linked_list */
  if (gv_first_MA == NULL){
    gv_first_MA = ma;
    gv_last_MA  = ma; 
    ma->prev    = NULL;
  }else{
    gv_last_MA->next = ma;
    ma->prev         = gv_last_MA;
    gv_last_MA       = ma;
  }
  ma->next      = NULL;

 
  /* keep the seeds chain inside MA */  
  KeepTupleList(tuple_list,
		first_tuple_aligned,
		last_tuple_aligned,
		ma);
  
  /* initialise ma */
  ma->left_pos_begin   = left_pos_begin;
  ma->right_pos_begin  = right_pos_begin;
  ma->left_pos_end     = left_pos_end;
  ma->right_pos_end    = right_pos_end;
  ma->blastscore       = blastscore;

  /*  
   * ma->mss              = mss;
   * ma->Mss              = Mss;
   * ma->nbs              = nbs; 
   */
  
  ma->reverse          = reverse;
  ma->i_chunk          = i_chunk;
  
 
  /* left_correction take effect on TL_POS(t): we can do alignments before calling the "left_correction"
   * function on each MA, while left_correction global variable is != 0, or either after the "left_correction()"
   * call provided that left_correction global variable IS = 0.
   */

  
#ifdef	STATS
  nb_ma++;
#endif
  
  /* correction due to the algorithm "Assemble_Double" over "Assemble_Single" */
  LeftCorrection_MA(ma);

  
  return ma;
}

#ifdef INLINE
inline
#endif
void FreeMA(MA * ma)
{
  tuple  * t_, * t_prev_;  
  t_ = ma->first_tuple;
  while(t_){
    t_prev_ = t_ ;
    t_      = t_->next;
    FreeTuple(t_prev_);
  }
  free(ma);
}

#ifdef INLINE
inline
#endif
int LeftCorrection_MA(MA * ma)
{
  tuple * t = ma->first_tuple;
  while(t){
    t->distance -= gv_left_correction;
    t = t->next;
  }
  return 0;
}

#define QS_CRITERION blastscore

int QuickSort_MAList(MA ** p_firstMA, MA * last_ma /* set to NULL at the begining */)
{     
  
  MA *        ma_pivot  = *(p_firstMA);
  MA *        ma_list   = ma_pivot ;
  MA *      ma_current  ;
  MA * ma_prev_current  = ma_pivot ;
  int valeur_pivot;
  int valeur;
  
  if (ma_pivot == last_ma)
    return 0;
  
  ma_current   = ma_pivot->next;
  valeur_pivot = ma_pivot->QS_CRITERION;
  
  if (ma_current == last_ma)
    return 0;
  
  
  do{ 
    
    valeur = ma_current->QS_CRITERION;
    if ( valeur > valeur_pivot){
      /* put in the head of the list */
      ma_prev_current->next = ma_current->next;
      ma_current->next      = ma_list;
      ma_list               = ma_current;
      /* MAJ */
      ma_current            = ma_prev_current->next;
    }else{
      /* MAJ */
      ma_current      = ma_current->next;
      ma_prev_current = ma_prev_current->next;
    }    
  }while(ma_current != last_ma);
  
 
  QuickSort_MAList(&(ma_list) , ma_pivot)  ;/* qsort first part */  
  QuickSort_MAList(&(ma_pivot->next), ma_current);/* qsort second part */
  
  /* set the begining of the list at its firtst element */
  *(p_firstMA) = ma_list;
   
  return 0;
}

/*
 * ListSort
 */

int ListSort_MAList(MA ** p_firstMA,MA ** p_lastMA /* var set to NULL*/,int mask /* 0x40000000 */)
{
  
  MA * ma_0     = NULL, * ma_1     = NULL;
  MA * ma_0_top = NULL, * ma_1_top = NULL;

  MA * _ma  = *(p_firstMA);

  while(_ma){
    if((_ma->QS_CRITERION)&mask){
      if (!ma_1)
	ma_1 = _ma;
      else
	ma_1_top->next = _ma;
      ma_1_top = _ma;
    }else{
      if (!ma_0)
	ma_0 = _ma;
      else
	ma_0_top->next = _ma;
      ma_0_top = _ma;
    }
    _ma = _ma->next;
  }
  
  mask>>=1;
  
  if (ma_0_top){
    ma_0_top->next = NULL;
    if (mask > 0)
      ListSort_MAList(&ma_0,&ma_0_top,mask);
  }  
  if (ma_1_top){
    ma_1_top->next = NULL;
    if (mask > 0)
      ListSort_MAList(&ma_1,&ma_1_top,mask);
  }
  

  /* chain the two sublists into one */
  if (ma_1_top){
    *(p_firstMA) = ma_1;
    if(ma_0_top){
      ma_1_top->next = ma_0;
      *(p_lastMA)    = ma_0_top;
    }else{
      *(p_lastMA)    = ma_1_top;
    }
  }else{
    *(p_firstMA) = ma_0;
    *(p_lastMA)   = ma_0_top;
  }

  return 0;    
}


/* 
 * Remove MA with Evalue over MaxEvalue 
 */

int Prune_MAList(MA ** p_firstMA, double MaxEvalue )
{
  MA * ma, * ma_prev;
  int minScore = MinScore(gp_k_blast,gp_lambda_blast,gp_querysize,gp_textsize,MaxEvalue);
  
  ma = *(p_firstMA);
  if (!ma)
      return 0;

  /* find the first element OK and set it to *(p_firstMA) */
  while ( ma->blastscore < minScore  ){       
    ma_prev = ma;
    *(p_firstMA) = ma = ma->next;
    FreeMA(ma_prev);
    if (!ma)   
      return 0;
  }

  ma_prev      = ma;
  ma  = ma_prev->next;
  
  /* continue with the others MAs */
  while(ma){
    if( ma->blastscore < minScore){
      ma_prev->next = ma->next;
      FreeMA(ma);
      ma = ma_prev->next;
    }else{
      ma_prev = ma;
      ma      = ma->next;
    }
  }
  
  return 0;
}

int EntropyFilter_MA(MA * ma,char * dataquery, char * datarevquery,char * datatext,int * datatext_start)
{
  int k_entropy = 3;
  return 
    Entropy(
	    (ma->reverse)?datarevquery:dataquery, 
	    ma->left_pos_begin , 
	    ma->left_pos_end -  ma->left_pos_begin,
	    k_entropy
	    ) < gp_entropy_min
    || 
    Entropy(
	    (gp_nbfiles == 1)?dataquery:datatext+datatext_start[ma->i_chunk], 
	    ma->right_pos_begin , 
	    ma->right_pos_end -  ma->right_pos_begin ,
	    k_entropy
	    ) < gp_entropy_min
    ;
  
}



/*
 * Entropy filter : remove low complexity regions
 */

int EntropyFilter_MAList(MA ** p_firstMA,char * dataquery, char * datarevquery,char * datatext,int * datatext_start){
  MA * ma, * ma_prev;
    
  ma = *(p_firstMA);
  if (!ma)
      return 0;
  
   /* find the first element OK and set it to *(p_firstMA) */
  while (EntropyFilter_MA(ma,dataquery,datarevquery,datatext,datatext_start)){       
    ma_prev = ma;
    *(p_firstMA) = ma = ma->next;
    FreeMA(ma_prev);
    if (!ma)   
      return 0;
  }

  ma_prev      = ma;
  ma  = ma_prev->next;

  /* continue with the others MAs */
  while(ma){
    if(EntropyFilter_MA(ma,dataquery,datarevquery,datatext,datatext_start)){
      ma_prev->next = ma->next;
      FreeMA(ma);
      ma = ma_prev->next;
    }else{
      ma_prev = ma;
      ma      = ma->next;
    }
  }
  return 0;
}




/*
 * Display for  debug ...
 */




int DisplayListTupleList(tuplelist * tuplelist)
{
  while (tuplelist){
    DisplayTupleList(tuplelist);
    tuplelist = tuplelist->next;
  }
  return 0;
}

#ifdef INLINE
 inline 
#endif
int BitWeightTupleList(tuplelist * tl)
{
  tuple * tuple;
  int bitweight = 0;
  if (tl->first_tuple){
    tuple = tl->first_tuple;
    while(tuple){
      bitweight += (tuple->bitweight);
      tuple = tuple->next;
    }
  }
  return bitweight;
}


int DisplayTupleList(tuplelist * tuplelist)
{
  tuple * tuple;
  if (tuplelist->first_tuple){
    tuple = tuplelist->first_tuple;
    while(tuple){
      printf("(distance=%i;tbl=%i|tbr=%i|tsize=%i)",tuple->distance,TBL_POS(tuple),TBR_POS(tuple),TSIZE(tuple));
      tuple = tuple->next;
    }
    printf("\n");
  }
  return 0;
}

int DisplayMA(MA * ma)
{
  tuple * tuple = ma->first_tuple;
  while(tuple){
    printf("(occurrence = %i \t distance = %i \t bitweight = %i \t size = %i )\n",
	   tuple->occurrence,
	   tuple->distance,
	   tuple->bitweight,
	   TSIZE(tuple));
    
    tuple = tuple->next;
  }
  printf("\n"); 
  return 0;
}

int DisplayMAList(MA * first_ma)
{
  MA * ma = first_ma;
  while(ma){
    DisplayMA(ma);
    ma = ma->next;
  }
  return 0;
}


