/*
 *  YASS 1.07
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifndef __TUPLE_H_
#define __TUPLE_H_


typedef struct _tuple{
  struct _tuple * next;
  int    occurrence;
  int    distance;
  int    size;
  int    bitweight;
  int    overlappedhits;
} tuple;

typedef struct _tuplelist {
  struct _tuplelist  * next ;
  struct _tuple * first_tuple;
} tuplelist;


typedef struct _MA {
  struct _MA    * next;
  struct _MA    * prev;
  struct _tuple * first_tuple;
  unsigned char reverse;

  /* chunk used on the second sequence */
  int i_chunk;

  /* position of memorized alignment */
  int left_pos_begin  ;
  int right_pos_begin ;
  int left_pos_end    ;
  int right_pos_end   ;

  /* blastscore */ 
  int  blastscore;
    
  /* statistics of the chain */
  /*
   * int mss;
   * int Mss;
   * int nbs;
   */ 
} MA;



/* (1)
 *
 *  TBL_POS(t1)  TEL_POS(t1)                            TBR_POS(t1)  TER_POS(t1) 
 *     |           |                                        |            |
 *     |           |                                        |            |
 *-----[=========#=]-----[====#=]---------------------------[==========#=][====#=]-------------
 *               |            |                                        |       | 
 *               |            |                                        |       | 
 *          TL_POS(t1)  TL_POS(t2)                              TR_POS(t1)  TR_POS(t2)
 *
 *	
 *
 *
 *
 *
 *        TSIZE(t1) 
 *     <----------->
 *     .           .
 *     .           .
 *-----[=========#=]-----[====#=]---------------------------[==========#=][====#=]-------------
 *     .           .     .      .                           .            ..      . 
 *     .           .     .      .                           .            ..      .                    
 *     .           .     .      .                           .            ..      .  
 *     .           <----->      .                           .            ..      .              
 *     .         TGAP_L(t1,t2)  .                           .           -><-     .                         
 *     .                        .                           .     TGAP_R(t1,t2)  .             
 *     .                        .                           .                    .       
 *     <------------------------>                           .                    .                                              
 *            TSEG_L(t1,t2)                                 <-------------------->
 *                                                               TSEG_R(t1,t2) 
 */


/* position on the text (R as right) */
#define TR_POS(t) ((t)->occurrence)
 
/* position on the query (L as left) 
 *   TL_POS should be modified according to "left_correction" 
 *   whith can be either 0 or keysize
 *
 */
#define TL_POS1(t) ((t)->occurrence - (t)->distance)
#define TL_POS(t)  ((t)->occurrence - (t)->distance + gv_left_correction) 


/* position begin "B" and end "E" on each Left and Rigth seed */

#define TBR_POS(t)  (TR_POS(t) - (t)->size) 
#define TER_POS(t)  (TR_POS(t) + gp_seed_span )

#define TBL_POS(t)  (TL_POS(t) - (t)->size)
#define TEL_POS(t)  (TL_POS(t) + gp_seed_span )


/* length */

/* length of one segment (t1<t2) between 2 seeds (including seeds size) */
#define TSEG_L(t1,t2) (TL_POS1(t2) - TL_POS1(t1) + (t1)->size + gp_seed_span)
#define TSEG_R(t1,t2) (TR_POS(t2)  - TR_POS(t1)  + (t1)->size + gp_seed_span)

/* length of one gap     (t1<t2) between 2 seeds (excluding seeds size) */
#define TGAP_L(t1,t2) (TL_POS1(t2) - TL_POS1(t1) - (t2)->size - gp_seed_span)
#define TGAP_R(t1,t2) (TR_POS(t2)  - TR_POS(t1)  - (t2)->size - gp_seed_span)

/* full seed size */
#define TSIZE(t)  ((t)->size + gp_seed_span)

tuple     * CreateTuple    (int occurrence, int distance);
tuplelist * CreateTupleList();
MA        * CreateMA(char * data_query  , int datasize_query ,
		     char * data_text   , int datasize_text  ,
		     int left_pos_begin , int right_pos_begin ,
		     int left_pos_end   , int right_pos_end   ,
		     tuplelist * tuple_list,
		     tuple * first_tuple_aligned,
		     tuple * last_tuple_aligned ,
		     int score ,
		     /* int mss, int Mss, int nbs,*/
		     int reverse,int i_chunk);
void        FreeTuple(tuple * t);
void        FreeMA(MA * ma);
void        FreeTupleList(tuplelist ** p_tl,tuplelist ** p_tl_prev); /* used only in "align" */
int         KeepTupleList(tuplelist * tl,tuple * tfirst, tuple * tlast, MA * ma); /* cut and paste of the interestion chain into ma */

int         LeftCorrection_MA(MA * ma);
int          EntropyFilter_MA(MA * ma,char * dataquery, char * datarevquery,char * datatext,int * datatext_start);

int             QuickSort_MAList(MA ** p_firstMA,MA *  lastMA   /* set to NULL at begin */);
int              ListSort_MAList(MA ** p_firstMA,MA ** p_lastMA /* set to NULL at begin */,int mask /*set to  0x40000000 */);
int                 Prune_MAList(MA ** p_firstMA, double MaxEvalue );
int         EntropyFilter_MAList(MA ** p_firstMA,char * dataquery, char * datarevquery,char * datatext,int * datatext_start); /* */

int DisplayListTupleList(tuplelist * tuplelist);
int DisplayTupleList(tuplelist * tuplelist);
int DisplayMA(MA * ma);
int DisplayMAList(MA * first_ma);

#endif

